# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy
import datetime
import logging
import os
import re
from collections import defaultdict

import dateutil.parser
import shapefile
from dateutil import tz
from dateutil.relativedelta import relativedelta
from shapely.geometry import shape
from shapely.ops import unary_union

from eodag.api.product.metadata_mapping import (
    DEFAULT_METADATA_MAPPING,
    format_metadata,
    get_metadata_path,
)
from eodag.utils import (
    dict_items_recursive_apply,
    format_dict_items,
    jsonpath_parse_dict_items,
    string_to_jsonpath,
    update_nested_dict,
)
from eodag.utils.exceptions import (
    MisconfiguredError,
    NoMatchingProductType,
    NotAvailableError,
    ValidationError,
)

logger = logging.getLogger("eodag.rest.stac")

default_min_date = "2015-01-01"


class StacCommon(object):
    """Stac common object

    :param url: Requested URL
    :type url: str
    :param stac_config: STAC configuration from stac.yml conf file
    :type stac_config: dict
    :param provider: Chosen provider
    :type provider: str
    :param eodag_api: EODAG python API instance
    :type eodag_api: :class:`eodag.api.core.EODataAccessGateway`
    :param root: (optional) API root
    :type root: str
    """

    def __init__(
        self, url, stac_config, provider, eodag_api, root="/", *args, **kwargs
    ):
        self.url = url.rstrip("/") if len(url) > 1 else url
        self.stac_config = stac_config
        self.provider = provider
        self.eodag_api = eodag_api
        self.root = root.rstrip("/") if len(root) > 1 else root

        self.data = {}

        if self.provider and self.eodag_api.get_preferred_provider() != self.provider:
            self.eodag_api.set_preferred_provider(self.provider)

    def update_data(self, data):
        """Updates data using given input STAC dict data

        :param data: Catalog data (parsed STAC dict)
        :type data: dict
        """
        self.data.update(data)

        # bbox: str to float
        if (
            "extent" in self.data.keys()
            and "spatial" in self.data["extent"].keys()
            and "bbox" in self.data["extent"]["spatial"].keys()
        ):
            for i, bbox in enumerate(self.data["extent"]["spatial"]["bbox"]):
                self.data["extent"]["spatial"]["bbox"][i] = [float(x) for x in bbox]
        # "None" values to None
        self.data = dict_items_recursive_apply(
            self.data, lambda k, v: None if v == "None" else v
        )

        # ids and titles as str
        self.data = dict_items_recursive_apply(
            self.data, lambda k, v: str(v) if k in ["title", "id"] else v
        )

        # remove \n in descriptions
        try:
            self.data = dict_items_recursive_apply(
                self.data,
                lambda k, v: v.replace("\n", " ") if k == "description" else v,
            )
        except AttributeError as e:
            logger.warning("Could not format description = %s" % e)

        # empty stac_extensions: "" to []
        if not self.data.get("stac_extensions", True):
            self.data["stac_extensions"] = []

    @staticmethod
    def get_stac_extension(url, stac_config, extension, **kwargs):
        """Parse STAC extension from config and return as dict

        :param url: Requested URL
        :type url: str
        :param stac_config: STAC configuration from stac.yml conf file
        :type stac_config: dict
        :param extension: Extension name
        :type extension: str
        :param kwargs: Additional variables needed for parsing extension
        :type kwargs: Any
        :returns: STAC extension as dictionnary
        :rtype: dict
        """
        extension_model = (
            copy.deepcopy(stac_config).get("extensions", {}).get(extension, {})
        )

        # parse f-strings
        format_args = copy.deepcopy(stac_config)
        format_args["extension"] = {
            "url": url,
            "properties": kwargs.get("properties", {}),
        }
        extension = format_dict_items(extension_model, **format_args)

        return extension

    def as_dict(self):
        """Returns object data as dictionnary

        :returns: STAC data dictionnary
        :rtype: dict
        """
        return self.data

    __geo_interface__ = property(as_dict)


class StacItem(StacCommon):
    """Stac item object

    :param url: Requested URL
    :type url: str
    :param stac_config: STAC configuration from stac.yml conf file
    :type stac_config: dict
    :param provider: Chosen provider
    :type provider: str
    :param eodag_api: EODAG python API instance
    :type eodag_api: :class:`eodag.api.core.EODataAccessGateway`
    :param root: (optional) API root
    :type root: str
    """

    def __init__(
        self, url, stac_config, provider, eodag_api, root="/", *args, **kwargs
    ):
        super(StacItem, self).__init__(
            url=url,
            stac_config=stac_config,
            provider=provider,
            eodag_api=eodag_api,
            root=root,
            *args,
            **kwargs,
        )

    def __get_item_list(self, search_results, catalog):
        """Build STAC items list from EODAG search results

        :param search_results: EODAG search results
        :type search_results: :class:`~eodag.api.search_result.SearchResult`
        :param catalog: STAC catalog dict used for parsing item metadata
        :type catalog: dict
        :returns: STAC item dicts list
        :rtype: list
        """
        if len(search_results) <= 0:
            return []

        item_model = self.__filter_item_model_properties(
            self.stac_config["item"], search_results[0].product_type
        )

        # check if some items need to be converted
        need_conversion = {}
        for k, v in item_model["properties"].items():
            if isinstance(v, str):
                conversion, item_model["properties"][k] = get_metadata_path(
                    item_model["properties"][k]
                )
                if conversion is not None:
                    need_conversion[k] = conversion
                    # convert str to jsonpath if needed
                    item_model["properties"][k] = string_to_jsonpath(
                        k, item_model["properties"][k]
                    )

        item_list = []
        for product in search_results:
            # parse jsonpath
            product_item = jsonpath_parse_dict_items(
                item_model, {"product": product.__dict__}
            )
            # apply conversion if needed
            for prop_key, prop_val in need_conversion.items():
                conv_func, conv_args = prop_val
                # colon `:` in key breaks format() method, hide it
                formatable_prop_key = prop_key.replace(":", "")
                if conv_args is not None:
                    product_item["properties"][prop_key] = format_metadata(
                        "{%s#%s(%s)}" % (formatable_prop_key, conv_func, conv_args),
                        **{formatable_prop_key: product_item["properties"][prop_key]},
                    )
                else:
                    product_item["properties"][prop_key] = format_metadata(
                        "{%s#%s}" % (formatable_prop_key, conv_func),
                        **{formatable_prop_key: product_item["properties"][prop_key]},
                    )

            # parse f-strings
            format_args = copy.deepcopy(self.stac_config)
            format_args["catalog"] = catalog
            format_args["item"] = product_item
            product_item = format_dict_items(product_item, **format_args)
            product_item["bbox"] = [float(i) for i in product_item["bbox"]]

            # remove empty properties
            product_item = self.__filter_item_properties_values(product_item)

            item_list.append(product_item)

        return item_list

    def get_stac_items(self, search_results, catalog):
        """Build STAC items from EODAG search results

        :param search_results: EODAG search results
        :type search_results: :class:`~eodag.api.search_result.SearchResult`
        :param catalog: STAC catalog dict used for parsing item metadata
        :type catalog: dict
        :returns: Items dictionnary
        :rtype: dict
        """
        items_model = copy.deepcopy(self.stac_config["items"])

        search_results.numberMatched = search_results.properties["totalResults"]
        search_results.numberReturned = len(search_results)

        # next page link
        if "?" in self.url:
            # search endpoint: use page url as self link
            for i, _ in enumerate(items_model["links"]):
                if items_model["links"][i]["rel"] == "self":
                    items_model["links"][i]["href"] = catalog["url"]
            if "page=" not in self.url:
                search_results.next = "%s&page=%s" % (
                    self.url,
                    search_results.properties["page"] + 1,
                )
            else:
                search_results.next = re.sub(
                    r"^(.*)(page=[0-9]+)(.*)$",
                    r"\1page=%s\3" % (search_results.properties["page"] + 1),
                    self.url,
                )
        else:
            search_results.next = "%s?page=%s" % (
                self.url,
                search_results.properties["page"] + 1,
            )

        search_results.timeStamp = (
            datetime.datetime.now(datetime.timezone.utc)
            .isoformat()
            .replace("+00:00", "")
            + "Z"
        )

        # parse jsonpath
        items = jsonpath_parse_dict_items(
            items_model, {"search_results": search_results.__dict__}
        )
        # parse f-strings
        format_args = copy.deepcopy(self.stac_config)
        format_args["catalog"] = catalog
        items = format_dict_items(items, **format_args)

        # last page: remove next page link
        if (
            search_results.properties["itemsPerPage"]
            * search_results.properties["page"]
            >= search_results.properties["totalResults"]
        ):
            items["links"] = [link for link in items["links"] if link["rel"] != "next"]

        # provide static catalog to build features
        if "search?" in catalog["url"]:
            catalog["url"] = os.path.join(
                catalog["url"].split("search?")[0],
                "collections",
                catalog["id"],
            )
        else:
            catalog["url"] = catalog["url"].split("?")[0]
        items["features"] = self.__get_item_list(search_results, catalog)

        self.update_data(items)
        return self.as_dict()

    def __filter_item_model_properties(self, item_model, product_type):
        """Filter item model depending on product type metadata and its extensions.
        Removes not needed parameters, and adds supplementary ones as
        part of oseo extension.

        :param item_model: Item model from stac_config
        :type item_model: dict
        :param product_type: Product type
        :type product_type: str
        :returns: Filtered item model
        :rtype: dict
        """
        try:
            product_type_dict = [
                pt
                for pt in self.eodag_api.list_product_types(
                    provider=self.provider, fetch_providers=False
                )
                if pt["ID"] == product_type
            ][0]
        except IndexError:
            raise MisconfiguredError(
                "Product type {} not available for {}".format(
                    product_type, self.provider
                )
            )

        result_item_model = copy.deepcopy(item_model)
        result_item_model["stac_extensions"] = list(
            self.stac_config["stac_extensions"].values()
        )

        # build jsonpath for eodag product default properties and adapt path
        eodag_properties_dict = {
            k: string_to_jsonpath(k, v.replace("$.", "$.product."))
            for k, v in DEFAULT_METADATA_MAPPING.items()
            if "$.properties." in v
        }
        # add missing properties as oseo:missingProperty
        for k, v in eodag_properties_dict.items():
            if (
                v not in result_item_model["properties"].values()
                and k not in self.stac_config["metadata_ignore"]
            ):
                result_item_model["properties"]["oseo:" + k] = string_to_jsonpath(k, v)

        # Filter out unneeded extensions
        if product_type_dict["sensorType"] != "RADAR":
            result_item_model["stac_extensions"].remove(
                self.stac_config["stac_extensions"]["sar"]
            )

        # Filter out unneeded properties
        extensions_prefixes = [
            k
            for k, v in self.stac_config["stac_extensions"].items()
            if v in result_item_model["stac_extensions"]
        ]
        for k, v in item_model["properties"].items():
            # remove key if extension not in stac_extensions
            if ":" in k and k.split(":")[0] not in extensions_prefixes:
                result_item_model["properties"].pop(k, None)

        return result_item_model

    def __filter_item_properties_values(self, item):
        """Removes empty properties, unused extensions, and add missing extensions

        :param item: STAC item data
        :type item: dict
        :returns: Filtered item model
        :rtype: dict
        """
        all_extensions_dict = copy.deepcopy(self.stac_config["stac_extensions"])
        # parse f-strings with root
        all_extensions_dict = format_dict_items(
            all_extensions_dict, **{"catalog": {"root": self.root}}
        )

        item["stac_extensions"] = []
        # dict to list of keys to permit pop() while iterating
        for k in list(item["properties"]):
            extension_prefix = k.split(":")[0]

            if item["properties"][k] is None:
                item["properties"].pop(k, None)
            # feed found extensions list
            elif (
                extension_prefix in all_extensions_dict.keys()
                and all_extensions_dict[extension_prefix] not in item["stac_extensions"]
            ):
                # append path from item extensions
                item["stac_extensions"].append(all_extensions_dict[extension_prefix])

        return item

    def get_stac_item_from_product(self, product):
        """Build STAC item from EODAG product

        :param product: EODAG product
        :type product: :class:`eodag.api.product._product.EOProduct`
        :returns: STAC item
        :rtype: list
        """
        product_type = product.product_type

        item_model = self.__filter_item_model_properties(
            self.stac_config["item"], product_type
        )

        catalog = StacCatalog(
            url=self.url.split("/items")[0],
            stac_config=self.stac_config,
            root=self.root,
            provider=self.provider,
            eodag_api=self.eodag_api,
            catalogs=[product_type],
        )

        # parse jsonpath
        product_item = jsonpath_parse_dict_items(
            item_model, {"product": product.__dict__}
        )
        # parse f-strings
        format_args = copy.deepcopy(self.stac_config)
        # format_args["collection"] = dict(catalog.as_dict(), **{"url": catalog.url})
        format_args["catalog"] = dict(
            catalog.as_dict(), **{"url": catalog.url, "root": catalog.root}
        )
        format_args["item"] = product_item
        product_item = format_dict_items(product_item, **format_args)
        product_item["bbox"] = [float(i) for i in product_item["bbox"]]

        # remove empty properties
        product_item = self.__filter_item_properties_values(product_item)

        self.update_data(product_item)
        return self.as_dict()


class StacCollection(StacCommon):
    """Stac collection object

    :param url: Requested URL
    :type url: str
    :param stac_config: STAC configuration from stac.yml conf file
    :type stac_config: dict
    :param provider: Chosen provider
    :type provider: str
    :param eodag_api: EODAG python API instance
    :type eodag_api: :class:`eodag.api.core.EODataAccessGateway`
    :param root: (optional) API root
    :type root: str
    """

    def __init__(
        self, url, stac_config, provider, eodag_api, root="/", *args, **kwargs
    ):
        super(StacCollection, self).__init__(
            url=url,
            stac_config=stac_config,
            provider=provider,
            eodag_api=eodag_api,
            root=root,
            *args,
            **kwargs,
        )

    def __get_product_types(self, filters=None):
        """Returns a list of supported product types

        :param filters: (optional) Additional filters for product types search
        :type filters: dict
        :returns: List of corresponding product types
        :rtype: list
        """
        if filters is None:
            filters = {}
        try:
            guessed_product_types = self.eodag_api.guess_product_type(
                instrument=filters.get("instrument"),
                platform=filters.get("platform"),
                platformSerialIdentifier=filters.get("platformSerialIdentifier"),
                sensorType=filters.get("sensorType"),
                processingLevel=filters.get("processingLevel"),
            )
        except NoMatchingProductType:
            guessed_product_types = []
        if guessed_product_types:
            product_types = [
                pt
                for pt in self.eodag_api.list_product_types(provider=self.provider)
                if pt["ID"] in guessed_product_types
            ]
        else:
            product_types = self.eodag_api.list_product_types(provider=self.provider)
        return product_types

    def __get_collection_list(self, filters=None):
        """Build STAC collections list

        :param filters: (optional) Additional filters for collections search
        :type filters: dict
        :returns: STAC collection dicts list
        :rtype: list
        """
        collection_model = copy.deepcopy(self.stac_config["collection"])

        product_types = self.__get_product_types(filters)

        collection_list = []
        for product_type in product_types:
            # get default provider for each product_type
            product_type_provider = (
                self.provider
                or next(
                    self.eodag_api._plugins_manager.get_search_plugins(
                        product_type=product_type["ID"]
                    )
                ).provider
            )
            #
            # if provider not in self.eodag_api.available_providers():
            #     for p in self.eodag_api._plugins_manager.get_search_plugins(product_type=product_type["ID"]):
            #         print(p.provider)

            # parse jsonpath
            product_type_collection = jsonpath_parse_dict_items(
                collection_model,
                {
                    "product_type": product_type,
                    "provider": self.eodag_api.providers_config[
                        product_type_provider
                    ].__dict__,
                },
            )
            # parse f-strings
            format_args = copy.deepcopy(self.stac_config)
            format_args["collection"] = dict(
                product_type_collection, **{"url": self.url, "root": self.root}
            )
            product_type_collection = format_dict_items(
                product_type_collection, **format_args
            )

            collection_list.append(product_type_collection)

        return collection_list

    def get_collections(self, filters=None):
        """Build STAC collections

        :param filters: (optional) Additional filters for collections search
        :type filters: dict
        :returns: Collections dictionnary
        :rtype: dict
        """
        collections = copy.deepcopy(self.stac_config["collections"])
        collections["collections"] = self.__get_collection_list(filters)

        collections["links"] += [
            {
                "rel": "child",
                "title": collec["id"],
                "href": [
                    link["href"] for link in collec["links"] if link["rel"] == "self"
                ][0],
            }
            for collec in collections["collections"]
        ]

        self.update_data(collections)
        return self.as_dict()

    def get_collection_by_id(self, collection_id):
        """Build STAC collection by its id

        :param collection_id: Product type as collection ID
        :type collection_id: str
        :returns: Collection dictionnary
        :rtype: dict
        """
        collection_list = self.__get_collection_list()

        try:
            collection = [c for c in collection_list if c["id"] == collection_id][0]
        except IndexError:
            raise NotAvailableError("%s collection not found" % collection_id)

        self.update_data(collection)
        return self.as_dict()


class StacCatalog(StacCommon):
    """Stac Catalog object

    :param url: Requested URL
    :type url: str
    :param stac_config: STAC configuration from stac.yml conf file
    :type stac_config: dict
    :param provider: Chosen provider
    :type provider: str
    :param eodag_api: EODAG python API instance
    :type eodag_api: :class:`eodag.api.core.EODataAccessGateway`
    :param root: (optional) API root
    :type root: str
    :param catalogs: (optional) Catalogs list
    :type catalogs: list
    :param fetch_providers: (optional) Whether to fetch providers for new product
                            types or not
    :type fetch_providers: bool
    """

    def __init__(
        self,
        url,
        stac_config,
        provider,
        eodag_api,
        root="/",
        catalogs=[],
        fetch_providers=True,
        *args,
        **kwargs
    ):
        super(StacCatalog, self).__init__(
            url=url,
            stac_config=stac_config,
            provider=provider,
            eodag_api=eodag_api,
            root=root,
            *args,
            **kwargs,
        )
        self.shp_location_config = eodag_api.locations_config
        self.search_args = {}

        self.data = {}
        self.children = []

        self.catalog_config = copy.deepcopy(stac_config["catalog"])

        self.__update_data_from_catalog_config({"model": {}})

        # expand links
        self.parent = "/".join(self.url.rstrip("/").split("/")[:-1])
        if self.parent != self.root:
            self.data["links"].append({"rel": "parent", "href": self.parent})
        if self.children:
            self.data["links"] += self.children

        # build catalog
        self.__build_stac_catalog(catalogs, fetch_providers=fetch_providers)

    def __update_data_from_catalog_config(self, catalog_config):
        """Updates configuration and data using given input catalog config

        :param catalog_config: Catalog config, from yml stac_config[catalogs]
        :type catalog_config: dict
        """
        model = catalog_config["model"]

        self.catalog_config = update_nested_dict(self.catalog_config, catalog_config)

        # parse f-strings
        # defaultdict usage will return "" for missing keys in format_args
        format_args = copy.deepcopy(self.stac_config)
        format_args["catalog"] = defaultdict(
            str, dict(model, **{"root": self.root, "url": self.url})
        )
        parsed_model = format_dict_items(self.catalog_config["model"], **format_args)

        self.update_data(parsed_model)

        return True

    def set_children(self, children=[]):
        """Set catalog children / links

        :param children: (optional) Children list
        :type children: list
        """
        self.children = children
        self.data["links"] = [
            link for link in self.data["links"] if link["rel"] != "child"
        ]
        self.data["links"] += children
        return True

    def set_stac_product_type_by_id(self, product_type, **kwargs):
        """Updates catalog with given product_type

        :param product_type: Product type
        :type product_type: str
        """
        collection = StacCollection(
            url=self.url,
            stac_config=self.stac_config,
            provider=self.provider,
            eodag_api=self.eodag_api,
            root=self.root,
        ).get_collection_by_id(product_type)

        cat_model = copy.deepcopy(self.stac_config["catalogs"]["product_type"]["model"])
        # parse f-strings
        format_args = copy.deepcopy(self.stac_config)
        format_args["catalog"] = defaultdict(str, **self.data)
        format_args["collection"] = collection
        try:
            parsed_dict = format_dict_items(cat_model, **format_args)
        except Exception:
            logger.error("Could not format product_type catalog")
            raise

        self.update_data(parsed_dict)

        # update search args
        self.search_args.update({"product_type": product_type})

        return parsed_dict

    # get / set dates filters -------------------------------------------------

    def get_stac_years_list(self, **kwargs):
        """Get catalog available years list

        :returns: Years list
        :rtype: list
        """
        extent_date_min, extent_date_max = self.get_datetime_extent()

        return list(range(extent_date_min.year, extent_date_max.year + 1))

    def get_stac_months_list(self, **kwargs):
        """Get catalog available months list

        :returns: Months list
        :rtype: list
        """
        extent_date_min, extent_date_max = self.get_datetime_extent()

        return list(
            range(
                extent_date_min.month,
                (extent_date_max - relativedelta(days=1)).month + 1,
            )
        )

    def get_stac_days_list(self, **kwargs):
        """Get catalog available days list

        :returns: Days list
        :rtype: list
        """
        extent_date_min, extent_date_max = self.get_datetime_extent()

        return list(
            range(
                extent_date_min.day, (extent_date_max - relativedelta(days=1)).day + 1
            )
        )

    def set_stac_year_by_id(self, year, **kwargs):
        """Updates and returns catalog with given year

        :param year: Year number
        :type year: str
        :returns: Updated catalog
        :rtype: dict
        """
        extent_date_min, extent_date_max = self.get_datetime_extent()

        datetime_min = max(
            [extent_date_min, dateutil.parser.parse("{}-01-01T00:00:00Z".format(year))]
        )
        datetime_max = min(
            [
                extent_date_max,
                dateutil.parser.parse("{}-01-01T00:00:00Z".format((year)))
                + relativedelta(years=1),
            ]
        )

        catalog_model = copy.deepcopy(self.stac_config["catalogs"]["year"]["model"])

        parsed_dict = self.set_stac_date(datetime_min, datetime_max, catalog_model)

        return parsed_dict

    def set_stac_month_by_id(self, month, **kwargs):
        """Updates and returns catalog with given month

        :param month: Month number
        :type month: str
        :returns: Updated catalog
        :rtype: dict
        """
        extent_date_min, extent_date_max = self.get_datetime_extent()
        year = extent_date_min.year

        datetime_min = max(
            [
                extent_date_min,
                dateutil.parser.parse("{}-{}-01T00:00:00Z".format(year, month)),
            ]
        )
        datetime_max = min(
            [
                extent_date_max,
                dateutil.parser.parse("{}-{}-01T00:00:00Z".format(year, month))
                + relativedelta(months=1),
            ]
        )

        catalog_model = copy.deepcopy(self.stac_config["catalogs"]["month"]["model"])

        parsed_dict = self.set_stac_date(datetime_min, datetime_max, catalog_model)

        return parsed_dict

    def set_stac_day_by_id(self, day, **kwargs):
        """Updates and returns catalog with given day

        :param day: Day number
        :type day: str
        :returns: Updated catalog
        :rtype: dict
        """
        extent_date_min, extent_date_max = self.get_datetime_extent()
        year = extent_date_min.year
        month = extent_date_min.month

        datetime_min = max(
            [
                extent_date_min,
                dateutil.parser.parse("{}-{}-{}T00:00:00Z".format(year, month, day)),
            ]
        )
        datetime_max = min(
            [
                extent_date_max,
                dateutil.parser.parse("{}-{}-{}T00:00:00Z".format(year, month, day))
                + relativedelta(days=1),
            ]
        )

        catalog_model = copy.deepcopy(self.stac_config["catalogs"]["day"]["model"])

        parsed_dict = self.set_stac_date(datetime_min, datetime_max, catalog_model)

        return parsed_dict

    def get_datetime_extent(self):
        """Returns catalog temporal extent as datetime objs

        :returns: Start & stop dates
        :rtype: tuple
        """
        extent_date_min = dateutil.parser.parse(default_min_date).replace(tzinfo=tz.UTC)
        extent_date_max = datetime.datetime.now(datetime.timezone.utc).replace(
            tzinfo=tz.UTC
        )
        for interval in self.data["extent"]["temporal"]["interval"]:
            extent_date_min_str, extent_date_max_str = interval
            # date min
            if extent_date_min_str:
                extent_date_min = max(
                    extent_date_min, dateutil.parser.parse(extent_date_min_str)
                )
            # date max
            if extent_date_max_str:
                extent_date_max = min(
                    extent_date_max, dateutil.parser.parse(extent_date_max_str)
                )

        return (
            extent_date_min.replace(tzinfo=tz.UTC),
            extent_date_max.replace(tzinfo=tz.UTC),
        )

    def set_stac_date(self, datetime_min, datetime_max, catalog_model):
        """Updates catalog data using given dates

        :param datetime_min: Date min of interval
        :type datetime_min: :class:`datetime.datetime`
        :param datetime_max: Date max of interval
        :type datetime_max: :class:`datetime.datetime`
        :param catalog_model: Catalog model to use, from yml stac_config[catalogs]
        :type catalog_model: dict
        :returns: Updated catalog
        :rtype: dict
        """
        # parse f-strings
        format_args = copy.deepcopy(self.stac_config)
        format_args["catalog"] = defaultdict(str, **self.data)
        format_args["date"] = defaultdict(
            str,
            {
                "year": datetime_min.year,
                "month": datetime_min.month,
                "day": datetime_min.day,
                "min": datetime_min.isoformat().replace("+00:00", "") + "Z",
                "max": datetime_max.isoformat().replace("+00:00", "") + "Z",
            },
        )
        parsed_dict = format_dict_items(catalog_model, **format_args)

        self.update_data(parsed_dict)

        # update search args
        self.search_args.update(
            {
                "dtstart": datetime_min.isoformat().split("T")[0],
                "dtend": datetime_max.isoformat().split("T")[0],
            }
        )
        return parsed_dict

    # get / set cloud_cover filter --------------------------------------------

    def get_stac_cloud_covers_list(self, **kwargs):
        """Get cloud_cover list

        :returns: cloud_cover list
        :rtype: list
        """
        return list(range(0, 101, 10))

    def set_stac_cloud_cover_by_id(self, cloud_cover, **kwargs):
        """Updates and returns catalog with given max cloud_cover

        :param cloud_cover: Cloud_cover number
        :type cloud_cover: str
        :returns: Updated catalog
        :rtype: dict
        """
        cat_model = copy.deepcopy(self.stac_config["catalogs"]["cloud_cover"]["model"])
        # parse f-strings
        format_args = copy.deepcopy(self.stac_config)
        format_args["catalog"] = defaultdict(str, **self.data)
        format_args["cloud_cover"] = cloud_cover
        parsed_dict = format_dict_items(cat_model, **format_args)

        self.update_data(parsed_dict)

        # update search args
        self.search_args.update({"cloudCover": cloud_cover})

        return parsed_dict

    # get / set locations filter ----------------------------------------------

    def get_stac_location_list(self, catalog_name):
        """Get locations list using stac_conf & locations_config

        :param catalog_name: Catalog/location name
        :type catalog_name: str
        :returns: Locations list
        :rtype: list
        """

        if catalog_name not in self.stac_config["catalogs"]:
            logger.warning(
                "no entry found for {} in location_config".format(catalog_name)
            )
            return []
        location_config = self.stac_config["catalogs"][catalog_name]

        for k in ["path", "attr"]:
            if k not in location_config.keys():
                logger.warning(
                    "no {} key found for {} in location_config".format(k, catalog_name)
                )
                return []
        path = location_config["path"]
        attr = location_config["attr"]

        with shapefile.Reader(path) as shp:
            countries_list = [rec[attr] for rec in shp.records()]

        # remove duplicates
        countries_list = list(set(countries_list))

        countries_list.sort()

        return countries_list

    def set_stac_location_by_id(self, location, catalog_name):
        """Updates and returns catalog with given location

        :param location: Feature attribute value for shp filtering
        :type location: str
        :param catalog_name: Catalog/location name
        :type catalog_name: str
        :returns: Updated catalog
        :rtype: dict
        """
        location_list_cat_key = catalog_name + "_list"

        if location_list_cat_key not in self.stac_config["catalogs"]:
            logger.warning(
                "no entry found for {}'s list in location_config".format(catalog_name)
            )
            return {}
        location_config = self.stac_config["catalogs"][location_list_cat_key]

        for k in ["path", "attr"]:
            if k not in location_config.keys():
                logger.warning(
                    "no {} key found for {}'s list in location_config".format(
                        k, catalog_name
                    )
                )
                return {}
        path = location_config["path"]
        attr = location_config["attr"]

        with shapefile.Reader(path) as shp:
            geom_hits = [
                shape(shaperec.shape)
                for shaperec in shp.shapeRecords()
                if shaperec.record.as_dict().get(attr, None) == location
            ]

        if len(geom_hits) == 0:
            logger.warning(
                "no feature found in %s matching %s=%s" % (path, attr, location)
            )
            return {}

        geom = unary_union(geom_hits)

        cat_model = copy.deepcopy(self.stac_config["catalogs"]["country"]["model"])
        # parse f-strings
        format_args = copy.deepcopy(self.stac_config)
        format_args["catalog"] = defaultdict(str, **self.data)
        format_args["feature"] = defaultdict(str, {"geometry": geom, "id": location})
        parsed_dict = format_dict_items(cat_model, **format_args)

        self.update_data(parsed_dict)

        # update search args
        self.search_args.update({"geom": geom})

        return parsed_dict

    def build_locations_config(self):
        """Build locations config from stac_conf[locations_catalogs] & eodag_api.locations_config

        :returns: Locations configuration dict
        :rtype: dict
        """
        user_config_locations_list = self.eodag_api.locations_config

        locations_config_model = copy.deepcopy(self.stac_config["locations_catalogs"])

        locations_config = {}
        for loc in user_config_locations_list:
            # parse jsonpath
            parsed = jsonpath_parse_dict_items(
                locations_config_model, {"shp_location": loc}
            )

            # set default child/parent for this location
            parsed["location"]["parent_key"] = "{}_list".format(loc["name"])

            locations_config["{}_list".format(loc["name"])] = parsed["locations_list"]
            locations_config[loc["name"]] = parsed["location"]

        return locations_config

    def __build_stac_catalog(self, catalogs=[], fetch_providers=True):
        """Build nested catalog from catalag list

        :param catalogs: (optional) Catalogs list
        :type catalogs: list
        :param fetch_providers: (optional) Whether to fetch providers for new product
                                types or not
        :type fetch_providers: bool
        :returns: This catalog obj
        :rtype: :class:`eodag.stac.StacCatalog`
        """
        # update conf with user shp locations
        locations_config = self.build_locations_config()

        self.stac_config["catalogs"] = dict(
            copy.deepcopy(self.stac_config["catalogs"]), **locations_config
        )

        if len(catalogs) == 0:
            # Build root catalog combined with landing page
            self.__update_data_from_catalog_config(
                {
                    "model": dict(
                        copy.deepcopy(self.stac_config["landing_page"]),
                        **{"provider": self.provider},
                    )
                }
                # {"model": copy.deepcopy(self.stac_config["landing_page"])}
            )

            # build children : product_types
            product_types_list = [
                pt
                for pt in self.eodag_api.list_product_types(
                    provider=self.provider, fetch_providers=fetch_providers
                )
            ]
            self.set_children(
                [
                    {
                        "rel": "child",
                        "href": self.url + "/" + product_type["ID"],
                        "title": product_type["ID"],
                    }
                    for product_type in product_types_list
                ]
            )
        else:
            # use product_types_list as base for building nested catalogs
            self.__update_data_from_catalog_config(
                copy.deepcopy(self.stac_config["catalogs"]["product_types_list"])
            )

        for idx, cat in enumerate(catalogs):
            if idx % 2 == 0:
                # even: cat is a filtering value ----------------------------------
                cat_data_name = self.catalog_config["child_key"]
                cat_data_value = cat

                # update data
                set_data_method_name = (
                    "set_stac_%s_by_id" % cat_data_name
                    if "catalog_type"
                    not in self.stac_config["catalogs"][cat_data_name].keys()
                    else "set_stac_%s_by_id"
                    % self.stac_config["catalogs"][cat_data_name]["catalog_type"]
                )
                set_data_method = getattr(self, set_data_method_name)
                set_data_method(cat_data_value, catalog_name=cat_data_name)

                if idx == len(catalogs) - 1:
                    # build children : remaining filtering keys
                    remaining_catalogs_list = [
                        c
                        for c in self.stac_config["catalogs"].keys()
                        # keep filters not used yet AND
                        if self.stac_config["catalogs"][c]["model"]["id"]
                        not in catalogs
                        and (
                            # filters with no parent_key constraint (no key, or key=None) OR
                            "parent_key" not in self.stac_config["catalogs"][c]
                            or not self.stac_config["catalogs"][c]["parent_key"]
                            # filters matching parent_key constraint
                            or self.stac_config["catalogs"][c]["parent_key"]
                            == cat_data_name
                        )
                        # AND filters that match parent attr constraint (locations)
                        and (
                            "parent" not in self.stac_config["catalogs"][c]
                            or not self.stac_config["catalogs"][c]["parent"]["key"]
                            or (
                                self.stac_config["catalogs"][c]["parent"]["key"]
                                == cat_data_name
                                and self.stac_config["catalogs"][c]["parent"]["attr"]
                                == cat_data_value
                            )
                        )
                    ]

                    self.set_children(
                        [
                            {
                                "rel": "child",
                                "href": self.url
                                + "/"
                                + self.stac_config["catalogs"][c]["model"]["id"],
                                "title": str(
                                    self.stac_config["catalogs"][c]["model"]["id"]
                                ),
                            }
                            for c in remaining_catalogs_list
                        ]
                        + [
                            {
                                "rel": "items",
                                "href": self.url + "/items",
                                "title": "items",
                            }
                        ]
                    )

            else:
                # odd: cat is a filtering key -------------------------------------
                try:
                    cat_key = [
                        c
                        for c in self.stac_config["catalogs"].keys()
                        if self.stac_config["catalogs"][c]["model"]["id"] == cat
                    ][0]
                except IndexError:
                    raise ValidationError(
                        "Bad settings for %s in stac_config catalogs" % cat
                    )
                cat_config = copy.deepcopy(self.stac_config["catalogs"][cat_key])
                # update data
                self.__update_data_from_catalog_config(cat_config)

                # get filtering values list
                get_data_method_name = (
                    "get_stac_%s" % cat_key
                    if "catalog_type"
                    not in self.stac_config["catalogs"][cat_key].keys()
                    else "get_stac_%s"
                    % self.stac_config["catalogs"][cat_key]["catalog_type"]
                )
                get_data_method = getattr(self, get_data_method_name)
                cat_data_list = get_data_method(catalog_name=cat_key)

                if idx == len(catalogs) - 1:
                    # filtering values list as children (do not include items)
                    self.set_children(
                        [
                            {
                                "rel": "child",
                                "href": self.url + "/" + str(filtering_data),
                                "title": str(filtering_data),
                            }
                            for filtering_data in cat_data_list
                        ]
                    )

        return self

    def get_stac_catalog(self):
        """Get nested STAC catalog as data dict

        :returns: Catalog dictionnary
        :rtype: dict
        """
        return self.as_dict()



# -*- coding: utf-8 -*-
# Copyright 2018, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy
import logging
import shutil
import tarfile
import zipfile

import requests
from jsonpath_ng.ext import parse
from requests import RequestException
from usgs import USGSAuthExpiredError, USGSError, api

from eodag.api.product import EOProduct
from eodag.api.product.metadata_mapping import (
    DEFAULT_METADATA_MAPPING,
    mtd_cfg_as_jsonpath,
    properties_from_json,
)
from eodag.plugins.apis.base import Api
from eodag.plugins.download.base import (
    DEFAULT_DOWNLOAD_TIMEOUT,
    DEFAULT_DOWNLOAD_WAIT,
    Download,
)
from eodag.utils import (
    GENERIC_PRODUCT_TYPE,
    ProgressCallback,
    format_dict_items,
    path_to_uri,
)
from eodag.utils.exceptions import AuthenticationError, NotAvailableError

logger = logging.getLogger("eodag.plugins.apis.usgs")


class UsgsApi(Download, Api):
    """A plugin that enables to query and download data on the USGS catalogues"""

    def authenticate(self):
        """Login to usgs api

        :raises: :class:`~eodag.utils.exceptions.AuthenticationError`
        """
        for i in range(2):
            try:
                api.login(
                    getattr(self.config, "credentials", {}).get("username", ""),
                    getattr(self.config, "credentials", {}).get("password", ""),
                    save=True,
                )
                break
            # if API key expired, retry to login once after logout
            except USGSAuthExpiredError:
                api.logout()
                continue
            except USGSError:
                raise AuthenticationError(
                    "Please check your USGS credentials."
                ) from None

    def query(
        self, product_type=None, items_per_page=None, page=None, count=True, **kwargs
    ):
        """Search for data on USGS catalogues"""
        product_type = kwargs.get("productType")
        if product_type is None:
            return [], 0

        self.authenticate()

        product_type_def_params = self.config.products.get(
            product_type, self.config.products[GENERIC_PRODUCT_TYPE]
        )
        usgs_dataset = format_dict_items(product_type_def_params, **kwargs)["dataset"]
        start_date = kwargs.pop("startTimeFromAscendingNode", None)
        end_date = kwargs.pop("completionTimeFromAscendingNode", None)
        geom = kwargs.pop("geometry", None)
        footprint = {}
        if hasattr(geom, "bounds"):
            (
                footprint["lonmin"],
                footprint["latmin"],
                footprint["lonmax"],
                footprint["latmax"],
            ) = geom.bounds
        else:
            footprint = geom

        final = []
        if footprint and len(footprint.keys()) == 4:  # a rectangle (or bbox)
            lower_left = {
                "longitude": footprint["lonmin"],
                "latitude": footprint["latmin"],
            }
            upper_right = {
                "longitude": footprint["lonmax"],
                "latitude": footprint["latmax"],
            }
        else:
            lower_left, upper_right = None, None
        try:
            api_search_kwargs = dict(
                start_date=start_date,
                end_date=end_date,
                ll=lower_left,
                ur=upper_right,
                max_results=items_per_page,
                starting_number=(1 + (page - 1) * items_per_page),
            )
            logger.info(
                f"Sending search request for {usgs_dataset} with {api_search_kwargs}"
            )

            results = api.scene_search(usgs_dataset, **api_search_kwargs)

            # update results with storage info from download_options()
            results_by_entity_id = {
                res["entityId"]: res for res in results["data"]["results"]
            }
            logger.debug(
                f"Adapting {len(results_by_entity_id)} plugin results to eodag product representation"
            )
            download_options = api.download_options(
                usgs_dataset, list(results_by_entity_id.keys())
            )
            if download_options.get("data", None) is not None:
                for download_option in download_options["data"]:
                    if "dds" in download_option["downloadSystem"]:
                        results_by_entity_id[download_option["entityId"]].update(
                            download_option
                        )
            results["data"]["results"] = list(results_by_entity_id.values())

            # Same method as in base.py, Search.__init__()
            # Prepare the metadata mapping
            # Do a shallow copy, the structure is flat enough for this to be sufficient
            metas = DEFAULT_METADATA_MAPPING.copy()
            # Update the defaults with the mapping value. This will add any new key
            # added by the provider mapping that is not in the default metadata.
            # A deepcopy is done to prevent self.config.metadata_mapping from being modified when metas[metadata]
            # is a list and is modified
            metas.update(copy.deepcopy(self.config.metadata_mapping))
            metas = mtd_cfg_as_jsonpath(metas)

            for result in results["data"]["results"]:

                result["productType"] = usgs_dataset

                product_properties = properties_from_json(result, metas)

                final.append(
                    EOProduct(
                        productType=product_type,
                        provider=self.provider,
                        properties=product_properties,
                        geometry=footprint,
                    )
                )
        except USGSError as e:
            logger.warning(
                f"Product type {usgs_dataset} does not exist on USGS EE catalog"
            )
            logger.warning(f"Skipping error: {e}")
        api.logout()

        if final:
            # parse total_results
            path_parsed = parse(self.config.pagination["total_items_nb_key_path"])
            total_results = path_parsed.find(results["data"])[0].value
        else:
            total_results = 0

        return final, total_results

    def download(
        self,
        product,
        auth=None,
        progress_callback=None,
        wait=DEFAULT_DOWNLOAD_WAIT,
        timeout=DEFAULT_DOWNLOAD_TIMEOUT,
        **kwargs,
    ):
        """Download data from USGS catalogues"""

        if progress_callback is None:
            logger.info(
                "Progress bar unavailable, please call product.download() instead of plugin.download()"
            )
            progress_callback = ProgressCallback(disable=True)

        outputs_extension = self.config.products.get(
            product.product_type, self.config.products[GENERIC_PRODUCT_TYPE]
        ).get("outputs_extension", ".tar.gz")

        fs_path, record_filename = self._prepare_download(
            product,
            progress_callback=progress_callback,
            outputs_extension=outputs_extension,
            **kwargs,
        )
        if not fs_path or not record_filename:
            if fs_path:
                product.location = path_to_uri(fs_path)
            return fs_path

        self.authenticate()

        if "dds" in product.properties.get("downloadSystem", ""):
            raise NotAvailableError(
                f"No USGS products found for {product.properties['id']}"
            )

        download_request = api.download_request(
            product.properties["productType"],
            product.properties["entityId"],
            product.properties["productId"],
        )

        req_urls = []
        try:
            if len(download_request["data"]["preparingDownloads"]) > 0:
                req_urls.extend(
                    [x["url"] for x in download_request["data"]["preparingDownloads"]]
                )
            else:
                req_urls.extend(
                    [x["url"] for x in download_request["data"]["availableDownloads"]]
                )
        except KeyError as e:
            raise NotAvailableError(
                f"{e} not found in {product.properties['id']} download_request"
            )

        if len(req_urls) > 1:
            logger.warning(
                f"{len(req_urls)} usgs products found for {product.properties['id']}. Only first will be downloaded"
            )
        elif not req_urls:
            raise NotAvailableError(
                f"No usgs request url was found for {product.properties['id']}"
            )

        req_url = req_urls[0]
        progress_callback.reset()
        logger.debug(f"Downloading {req_url}")

        @self._download_retry(product, wait, timeout)
        def download_request(product, fs_path, progress_callback, **kwargs):
            try:
                with requests.get(
                    req_url,
                    stream=True,
                    timeout=wait * 60,
                ) as stream:
                    try:
                        stream.raise_for_status()
                    except RequestException:
                        import traceback as tb

                        logger.error(
                            f"Error while getting resource :\n{tb.format_exc()}",
                        )
                    else:
                        stream_size = int(stream.headers.get("content-length", 0))
                        progress_callback.reset(total=stream_size)
                        with open(fs_path, "wb") as fhandle:
                            for chunk in stream.iter_content(chunk_size=64 * 1024):
                                if chunk:
                                    fhandle.write(chunk)
                                    progress_callback(len(chunk))
            except requests.exceptions.Timeout as e:
                raise NotAvailableError(str(e))

        download_request(product, fs_path, progress_callback, **kwargs)

        with open(record_filename, "w") as fh:
            fh.write(product.properties["downloadLink"])
        logger.debug(f"Download recorded in {record_filename}")

        api.logout()

        # Check downloaded file format
        if (outputs_extension == ".tar.gz" and tarfile.is_tarfile(fs_path)) or (
            outputs_extension == ".zip" and zipfile.is_zipfile(fs_path)
        ):
            product_path = self._finalize(
                fs_path,
                progress_callback=progress_callback,
                outputs_extension=outputs_extension,
                **kwargs,
            )
            product.location = path_to_uri(product_path)
            return product_path
        elif tarfile.is_tarfile(fs_path):
            logger.info(
                "Downloaded product detected as a tar File, but was was expected to be a zip file"
            )
            new_fs_path = fs_path[: fs_path.index(outputs_extension)] + ".tar.gz"
            shutil.move(fs_path, new_fs_path)
            product.location = path_to_uri(new_fs_path)
            return new_fs_path
        elif zipfile.is_zipfile(fs_path):
            logger.info(
                "Downloaded product detected as a zip File, but was was expected to be a tar file"
            )
            new_fs_path = fs_path[: fs_path.index(outputs_extension)] + ".zip"
            shutil.move(fs_path, new_fs_path)
            product.location = path_to_uri(new_fs_path)
            return new_fs_path
        else:
            logger.warning(
                "Downloaded product is not a tar or a zip File. Please check its file type before using it"
            )
            new_fs_path = fs_path[: fs_path.index(outputs_extension)]
            shutil.move(fs_path, new_fs_path)
            product.location = path_to_uri(new_fs_path)
            return new_fs_path

    def download_all(
        self,
        products,
        auth=None,
        downloaded_callback=None,
        progress_callback=None,
        wait=DEFAULT_DOWNLOAD_WAIT,
        timeout=DEFAULT_DOWNLOAD_TIMEOUT,
        **kwargs,
    ):
        """
        Download all using parent (base plugin) method
        """
        return super(UsgsApi, self).download_all(
            products,
            auth=auth,
            downloaded_callback=downloaded_callback,
            progress_callback=progress_callback,
            wait=wait,
            timeout=timeout,
            **kwargs,
        )



# -*- coding: utf-8 -*-
# Copyright 2018, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import io
import logging
import os
import sys
import traceback
from distutils import dist
from functools import wraps

import flask
import geojson
import pkg_resources
from flasgger import Swagger
from flask import abort, jsonify, make_response, request, send_file

from eodag.config import load_stac_api_config
from eodag.rest.utils import (  # get_stac_landing_page,; get_stac_product_types_catalog,; search_products,
    download_stac_item_by_id,
    get_detailled_collections_list,
    get_stac_catalogs,
    get_stac_collection_by_id,
    get_stac_collections,
    get_stac_conformance,
    get_stac_extension_oseo,
    get_stac_item_by_id,
    load_stac_config,
    search_stac_items,
)
from eodag.utils.exceptions import (
    MisconfiguredError,
    NoMatchingProductType,
    NotAvailableError,
    UnsupportedProductType,
    UnsupportedProvider,
    ValidationError,
)

logger = logging.getLogger("eodag.rest.server")

app = flask.Flask(__name__)

# eodag metadata
distribution = pkg_resources.get_distribution("eodag")
metadata_str = distribution.get_metadata(distribution.PKG_INFO)
metadata_obj = dist.DistributionMetadata()
metadata_obj.read_pkg_file(io.StringIO(metadata_str))

# confs from resources yaml files
stac_config = load_stac_config()

stac_api_config = {"info": {}}
stac_api_config = load_stac_api_config()
root_catalog = get_stac_catalogs(url="", fetch_providers=False)
stac_api_version = stac_api_config["info"]["version"]
stac_api_config["info"]["title"] = root_catalog["title"] + " / eodag"
stac_api_config["info"]["description"] = (
    root_catalog["description"]
    + " (stac-api-spec {})".format(stac_api_version)
    + "".join(
        [
            "\n - [{0}](/collections/{0}): {1}".format(pt["ID"], pt["abstract"])
            for pt in get_detailled_collections_list(fetch_providers=False)
        ]
    )
)
stac_api_config["info"]["version"] = getattr(
    metadata_obj, "version", stac_api_config["info"]["version"]
)
stac_api_config["info"]["contact"]["name"] = "EODAG"
stac_api_config["info"]["contact"]["url"] = getattr(
    metadata_obj, "url", stac_api_config["info"]["contact"]["url"]
)
stac_api_config["title"] = root_catalog["title"] + " - service-doc"
stac_api_config["specs"] = [
    {
        "endpoint": "service-desc",
        "route": "/api",
        "rule_filter": lambda rule: True,  # all in
        "model_filter": lambda tag: True,  # all in
    }
]
stac_api_config["static_url_path"] = "/service-static"
stac_api_config["specs_route"] = "/service-doc/"
stac_api_config.pop("servers", None)
swagger = Swagger(app, config=stac_api_config, merge=True)


def cross_origin(request_handler):
    """Wraps a view to relax the need for csrf token"""

    @wraps(request_handler)
    def wrapper(*args, **kwargs):
        resp = make_response(request_handler(*args, **kwargs))
        resp.headers["Access-Control-Allow-Origin"] = "*"
        return resp

    return wrapper


@app.errorhandler(MisconfiguredError)
@app.errorhandler(NoMatchingProductType)
@app.errorhandler(UnsupportedProductType)
@app.errorhandler(UnsupportedProvider)
@app.errorhandler(ValidationError)
def handle_invalid_usage(error):
    """Invalid usage [400] errors handle"""
    status_code = 400
    response = jsonify(
        {"code": status_code, "name": type(error).__name__, "description": str(error)}
    )
    response.status_code = status_code
    logger.warning(traceback.format_exc())
    return response


@app.errorhandler(RuntimeError)
@app.errorhandler(Exception)
def handle_internal_error(error):
    """Internal [500] errors handle"""
    status_code = 500
    response = jsonify(
        {"code": status_code, "name": type(error).__name__, "description": str(error)}
    )
    response.status_code = status_code
    logger.error(traceback.format_exc())
    return response


@app.errorhandler(NotAvailableError)
@app.errorhandler(404)
def handle_resource_not_found(e):
    """Not found [404] errors handle"""
    return jsonify(error=str(e)), 404


@app.route("/conformance", methods=["GET"])
@cross_origin
def conformance():
    """STAC conformance"""

    response = get_stac_conformance()

    return jsonify(response), 200


@app.route("/", methods=["GET"])
@cross_origin
def catalogs_root():
    """STAC catalogs root"""

    response = get_stac_catalogs(
        url=request.url.split("?")[0],
        root=request.url_root,
        catalogs=[],
        provider=request.args.to_dict().get("provider", None),
    )

    return jsonify(response), 200


@app.route("/<path:catalogs>", methods=["GET"])
@cross_origin
def stac_catalogs(catalogs):
    """Describe the given catalog and list available sub-catalogs
    ---
    tags:
      - Capabilities
    parameters:
      - name: catalogs
        in: path
        required: true
        description: |-
            The catalog's path.

            For a nested catalog, provide the root-related path to the catalog (for example `S2_MSI_L1C/year/2020`)
        schema:
            type: string
    responses:
        200:
            description: The catalog's description
            content:
                application/json:
                    schema:
                        $ref: '#/components/schemas/collection'
        '500':
            $ref: '#/components/responses/ServerError'
    """

    catalogs = catalogs.strip("/").split("/")
    response = get_stac_catalogs(
        url=request.url.split("?")[0],
        root=request.url_root,
        catalogs=catalogs,
        provider=request.args.to_dict().get("provider", None),
    )
    return jsonify(response), 200


@app.route("/<path:catalogs>/items", methods=["GET"])
@cross_origin
def stac_catalogs_items(catalogs):
    """Fetch catalog's features
    ---
    tags:
      - Data
    description: |-
        Fetch features in the given catalog provided with `catalogs`.
    parameters:
      - name: catalogs
        in: path
        required: true
        description: |-
            The path to the catalog that contains the requested features.

            For a nested catalog, provide the root-related path to the catalog (for example `S2_MSI_L1C/year/2020`)
        schema:
            type: string
      - $ref: '#/components/parameters/bbox'
      - $ref: '#/components/parameters/datetime'
      - $ref: '#/components/parameters/limit'
    responses:
        200:
            description: The list of items found for the given catalog.
            type: array
            content:
                application/json:
                    schema:
                        $ref: '#/components/schemas/itemCollection'
        '500':
            $ref: '#/components/responses/ServerError
    '"""

    catalogs = catalogs.strip("/").split("/")
    arguments = request.args.to_dict()
    provider = arguments.pop("provider", None)
    response = search_stac_items(
        url=request.url,
        arguments=arguments,
        root=request.url_root,
        catalogs=catalogs,
        provider=provider,
    )
    return app.response_class(
        response=geojson.dumps(response), status=200, mimetype="application/json"
    )


@app.route("/<path:catalogs>/items/<item_id>", methods=["GET"])
@cross_origin
def stac_catalogs_item(catalogs, item_id):
    """Fetch catalog's single features
    ---
    tags:
      - Data
    description: |-
        Fetch the feature with id `featureId` in the given catalog provided.
        with `catalogs`.
    parameters:
        - name: catalogs
          in: path
          required: true
          description: |-
                The path to the catalog that contains the requested feature.


                For a nested catalog, provide the root-related path to the catalog (for example `S2_MSI_L1C/year/2020`)
          schema:
                type: string
        - name: item_id
          in: path
          description: |-
            local identifier of a feature (for example `S2A_MSIL1C_20200805T104031_N0209_R008_T31TCJ_20200805T110310`)
          required: true
          schema:
              type: string
    responses:
        '200':
          $ref: '#/components/responses/Feature'
        '404':
          $ref: '#/components/responses/NotFound'
        '500':
          $ref: '#/components/responses/ServerError'
    """

    catalogs = catalogs.strip("/").split("/")
    response = get_stac_item_by_id(
        url=request.url.split("?")[0],
        item_id=item_id,
        root=request.url_root,
        catalogs=catalogs,
        provider=request.args.to_dict().get("provider", None),
    )

    if response:
        return app.response_class(
            response=geojson.dumps(response), status=200, mimetype="application/json"
        )
    else:
        abort(
            404,
            "No item found matching `{}` id in catalog `{}`".format(item_id, catalogs),
        )


@app.route("/<path:catalogs>/items/<item_id>/download", methods=["GET"])
@cross_origin
def stac_catalogs_item_download(catalogs, item_id):
    """STAC item local download"""

    catalogs = catalogs.strip("/").split("/")
    response = download_stac_item_by_id(
        catalogs=catalogs,
        item_id=item_id,
        provider=request.args.to_dict().get("provider", None),
    )
    filename = os.path.basename(response)

    return send_file(response, as_attachment=True, attachment_filename=filename)


@app.route("/collections/", methods=["GET"])
@app.route("/collections", methods=["GET"])
@cross_origin
def collections():
    """STAC collections

    Can be filtered using parameters: instrument, platform, platformSerialIdentifier, sensorType, processingLevel
    """
    arguments = request.args.to_dict()
    provider = arguments.pop("provider", None)
    response = get_stac_collections(
        url=request.url.split("?")[0],
        root=request.url_root,
        arguments=arguments,
        provider=provider,
    )

    return jsonify(response), 200


@app.route("/collections/<collection_id>", methods=["GET"])
@cross_origin
def collection_by_id(collection_id):
    """STAC collection by id"""

    response = get_stac_collection_by_id(
        url=request.url.split("?")[0],
        root=request.url_root,
        collection_id=collection_id,
        provider=request.args.to_dict().get("provider", None),
    )

    return jsonify(response), 200


@app.route("/collections/<collection_id>/items", methods=["GET"])
@cross_origin
def stac_collections_items(collection_id):
    """STAC collections items"""

    arguments = request.args.to_dict()
    provider = arguments.pop("provider", None)
    response = search_stac_items(
        url=request.url,
        arguments=arguments,
        root=request.url_root,
        provider=provider,
        catalogs=[collection_id],
    )
    return app.response_class(
        response=geojson.dumps(response), status=200, mimetype="application/json"
    )


@app.route("/search", methods=["GET", "POST"])
@cross_origin
def stac_search():
    """STAC collections items"""

    if request.get_json(silent=True, force=True):
        arguments = dict(request.args.to_dict(), **request.get_json(force=True))
    else:
        arguments = request.args.to_dict()

    provider = arguments.pop("provider", None)
    response = search_stac_items(
        url=request.url, arguments=arguments, root=request.url_root, provider=provider
    )
    return app.response_class(
        response=geojson.dumps(response), status=200, mimetype="application/json"
    )


@app.route("/extensions/oseo/json-schema/schema.json", methods=["GET"])
@cross_origin
def stac_extension_oseo():
    """STAC OGC / OpenSearch extension for EO"""

    response = get_stac_extension_oseo(url=request.url.split("?")[0])

    return app.response_class(
        response=geojson.dumps(response), status=200, mimetype="application/json"
    )


@app.route("/collections/<collection_id>/items/<item_id>", methods=["GET"])
@cross_origin
def stac_collections_item(collection_id, item_id):
    """STAC collection item by id"""

    response = get_stac_item_by_id(
        url=request.url.split("?")[0],
        item_id=item_id,
        root=request.url_root,
        catalogs=[collection_id],
        provider=request.args.to_dict().get("provider", None),
    )

    return app.response_class(
        response=geojson.dumps(response), status=200, mimetype="application/json"
    )


@app.route("/collections/<collection_id>/items/<item_id>/download", methods=["GET"])
@cross_origin
def stac_collections_item_download(collection_id, item_id):
    """STAC collection item local download"""

    response = download_stac_item_by_id(
        catalogs=[collection_id],
        item_id=item_id,
        provider=request.args.to_dict().get("provider", None),
    )
    filename = os.path.basename(response)

    return send_file(response, as_attachment=True, attachment_filename=filename)


def main():
    """Launch the server"""
    import argparse

    parser = argparse.ArgumentParser(
        description="""Script for starting EODAG server""", epilog=""""""
    )
    parser.add_argument(
        "-d", "--daemon", action="store_true", help="run in daemon mode"
    )
    parser.add_argument(
        "-a",
        "--all-addresses",
        action="store_true",
        help="run flask using IPv4 0.0.0.0 (all network interfaces), "
        "otherwise bind to 127.0.0.1 (localhost). "
        "This maybe necessary in systems that only run Flask",
    )
    args = parser.parse_args()

    if args.all_addresses:
        bind_host = "0.0.0.0"
    else:
        bind_host = "127.0.0.1"

    if args.daemon:
        pid = None
        try:
            pid = os.fork()
        except OSError as e:
            raise Exception("%s [%d]" % (e.strerror, e.errno))

        if pid == 0:
            os.setsid()
            app.run(threaded=True, host=bind_host)
        else:
            sys.exit(0)
    else:
        # For development
        app.run(debug=True, use_reloader=True)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Copyright 2022, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import hashlib
import logging
from datetime import datetime

import cdsapi
import requests

from eodag import EOProduct
from eodag.api.product.metadata_mapping import NOT_AVAILABLE, properties_from_json
from eodag.plugins.apis.base import Api
from eodag.plugins.download.base import (
    DEFAULT_DOWNLOAD_TIMEOUT,
    DEFAULT_DOWNLOAD_WAIT,
    Download,
)
from eodag.plugins.search.qssearch import QueryStringSearch
from eodag.utils import (
    datetime_range,
    get_geometry_from_various,
    parse_qsl,
    path_to_uri,
    urlsplit,
)
from eodag.utils.exceptions import (
    AuthenticationError,
    DownloadError,
    RequestError,
    ValidationError,
)
from eodag.utils.logging import get_logging_verbose

logger = logging.getLogger("eodag.plugins.apis.cds")

CDS_KNOWN_FORMATS = {"grib": "grib", "netcdf": "nc"}


class CdsApi(Download, Api, QueryStringSearch):
    """A plugin that enables to build download-request and download data on CDS API.

    Builds a single ready-to-download :class:`~eodag.api.product._product.EOProduct`
    during the search stage.

    This class inherits from :class:`~eodag.plugins.apis.base.Api` for compatibility,
    :class:`~eodag.plugins.download.base.Download` for download methods, and
    :class:`~eodag.plugins.search.qssearch.QueryStringSearch` for metadata-mapping and
    query build methods.
    """

    def __init__(self, provider, config):
        # init self.config.metadata_mapping using Search Base plugin
        super(QueryStringSearch, self).__init__(provider, config)

        # needed by QueryStringSearch.build_query_string / format_free_text_search
        self.config.__dict__.setdefault("free_text_search_operations", {})
        # needed for compatibility
        self.config.__dict__.setdefault("pagination", {})

    def query(
        self, product_type=None, items_per_page=None, page=None, count=True, **kwargs
    ):
        """Build ready-to-download SearchResult"""

        # check some mandatory parameters
        # start date
        if "startTimeFromAscendingNode" not in kwargs:
            raise ValidationError("Required start date is missing")
        # end date
        if "completionTimeFromAscendingNode" not in kwargs:
            raise ValidationError("Required end date is missing")

        product_type = kwargs.get("productType")

        # Map query args to properties
        product_properties = kwargs
        product_properties.pop("auth", None)
        # geometry
        search_geometry = product_properties.get("geometry", None) or [
            -180,
            -90,
            180,
            90,
        ]
        product_geometry = get_geometry_from_various(geometry=search_geometry)
        product_properties["geometry"] = product_geometry

        # add product_type specific properties from provider configuration (overriden by query args)
        product_properties = dict(
            self.config.products.get(product_type, {}), **product_properties
        )

        # properties mapped using provider configuration
        product_properties = properties_from_json(
            product_properties, self.config.metadata_mapping
        )

        # build query string & parameters dict using from available mapped properties
        product_available_properties = {
            k: v for (k, v) in product_properties.items() if v != NOT_AVAILABLE
        }
        qp, qs = self.build_query_string(product_type, **product_available_properties)

        # query hash, will be used to build a product id
        query_hash = hashlib.sha1(str(qs).encode("UTF-8")).hexdigest()

        # build product id
        if product_type is not None:
            id_prefix = product_type
        else:
            id_prefix = ("%s" % (qp.get("dataset", ""))).upper()

        product_id = "%s_%s_%s" % (
            id_prefix,
            qp["date"][0].split("/")[0].replace("-", ""),
            query_hash,
        )
        product_properties["id"] = product_properties["title"] = product_id

        # update downloadLink
        product_properties["downloadLink"] += f"?{qs}"

        product = EOProduct(
            provider=self.provider,
            productType=product_type,
            properties=product_properties,
        )
        # use product_type_config as default properties
        product.properties = dict(
            getattr(self.config, "product_type_config", {}), **product.properties
        )

        results_count = 1
        return [
            product,
        ], results_count

    def _get_cds_client(self, **auth_dict):
        """Returns cdsapi client."""
        # eodag logging info
        eodag_verbosity = get_logging_verbose()
        eodag_logger = logging.getLogger("eodag")

        client = cdsapi.Client(
            # disable cdsapi default logging and handle it on eodag side
            # until https://github.com/ecmwf/cdsapi/pull/47 is merged
            quiet=True,
            verify=True,
            **auth_dict,
        )

        if eodag_verbosity is None or eodag_verbosity == 1:
            client.logger.setLevel(logging.WARNING)
        if eodag_verbosity == 2:
            client.logger.setLevel(logging.INFO)
        elif eodag_verbosity == 3:
            client.logger.setLevel(logging.DEBUG)
        else:
            client.logger.setLevel(logging.WARNING)

        if eodag_logger.hasHandlers():
            client.logger.addHandler(eodag_logger.handlers[0])

        return client

    def authenticate(self):
        """Returns information needed for auth

        :returns: {key, url} dictionary
        :rtype: dict
        :raises: :class:`~eodag.utils.exceptions.AuthenticationError`
        :raises: :class:`~eodag.utils.exceptions.RequestError`
        """
        # Get credentials from eodag or using cds conf
        uid = getattr(self.config, "credentials", {}).get("username", None)
        api_key = getattr(self.config, "credentials", {}).get("password", None)
        url = getattr(self.config, "api_endpoint", None)
        if not all([uid, api_key, url]):
            raise AuthenticationError("Missing authentication informations")

        auth_dict = {"key": f"{uid}:{api_key}", "url": url}

        client = self._get_cds_client(**auth_dict)
        try:
            client.status()
            logger.debug("Connection checked on CDS API")
        except requests.exceptions.ConnectionError as e:
            logger.error(e)
            raise RequestError(f"Could not connect to the CDS API '{url}'")
        except requests.exceptions.HTTPError as e:
            logger.error(e)
            raise RequestError("The CDS API has returned an unexpected error")

        return auth_dict

    def download(self, product, auth=None, progress_callback=None, **kwargs):
        """Download data from providers using CDS API"""

        product_extension = CDS_KNOWN_FORMATS[product.properties.get("format", "grib")]

        # Prepare download
        fs_path, record_filename = self._prepare_download(
            product,
            progress_callback=progress_callback,
            outputs_extension=f".{product_extension}",
            **kwargs,
        )

        if not fs_path or not record_filename:
            if fs_path:
                product.location = path_to_uri(fs_path)
            return fs_path

        # get download request dict from product.location/downloadLink url query string
        query_str = "".join(urlsplit(product.location).fragment.split("?", 1)[1:])
        download_request = dict(parse_qsl(query_str))

        date_range = download_request.pop("date_range", False)
        if date_range:
            date = download_request.pop("date")
            start, end, *_ = date.split("/")
            _start = datetime.fromisoformat(start)
            _end = datetime.fromisoformat(end)
            d_range = [d for d in datetime_range(_start, _end)]
            download_request["year"] = [*{str(d.year) for d in d_range}]
            download_request["month"] = [*{str(d.month) for d in d_range}]
            download_request["day"] = [*{str(d.day) for d in d_range}]

        auth_dict = self.authenticate()
        dataset_name = download_request.pop("dataset")

        # Send download request to CDS web API
        logger.info(
            "Request download on CDS API: dataset=%s, request=%s",
            dataset_name,
            download_request,
        )
        try:
            client = self._get_cds_client(**auth_dict)
            client.retrieve(name=dataset_name, request=download_request, target=fs_path)
        except Exception as e:
            logger.error(e)
            raise DownloadError(e)

        with open(record_filename, "w") as fh:
            fh.write(product.properties["downloadLink"])
        logger.debug("Download recorded in %s", record_filename)

        # do not try to extract or delete grib/netcdf
        kwargs["extract"] = False

        product_path = self._finalize(
            fs_path,
            progress_callback=progress_callback,
            outputs_extension=f".{product_extension}",
            **kwargs,
        )
        product.location = path_to_uri(product_path)
        return product_path

    def download_all(
        self,
        products,
        auth=None,
        downloaded_callback=None,
        progress_callback=None,
        wait=DEFAULT_DOWNLOAD_WAIT,
        timeout=DEFAULT_DOWNLOAD_TIMEOUT,
        **kwargs,
    ):
        """
        Download all using parent (base plugin) method
        """
        return super(CdsApi, self).download_all(
            products,
            auth=auth,
            downloaded_callback=downloaded_callback,
            progress_callback=progress_callback,
            wait=wait,
            timeout=timeout,
            **kwargs,
        )



# -*- coding: utf-8 -*-
# Copyright 2018, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import re
import shutil
from pathlib import Path

import boto3
import requests
from botocore.exceptions import ClientError, ProfileNotFound
from botocore.handlers import disable_signing
from lxml import etree

from eodag.api.product.metadata_mapping import (
    mtd_cfg_as_jsonpath,
    properties_from_json,
    properties_from_xml,
)
from eodag.plugins.download.base import Download
from eodag.utils import ProgressCallback, path_to_uri, rename_subfolder, urlparse
from eodag.utils.exceptions import AuthenticationError, DownloadError
from eodag.utils.stac_reader import HTTP_REQ_TIMEOUT

logger = logging.getLogger("eodag.plugins.download.aws")

# AWS chunk path identify patterns

# S2 L2A Tile files -----------------------------------------------------------
S2L2A_TILE_IMG_REGEX = re.compile(
    r"^tiles/(?P<tile1>[0-9]+)/(?P<tile2>[A-Z]+)/(?P<tile3>[A-Z]+)/(?P<year>[0-9]+)/(?P<month>[0-9]+)/"
    + r"(?P<day>[0-9]+)/(?P<num>[0-9]+)/R(?P<res>[0-9]+m)/(?P<file>[A-Z0-9_]+)\.jp2$"
)
S2L2A_TILE_AUX_DIR_REGEX = re.compile(
    r"^tiles/(?P<tile1>[0-9]+)/(?P<tile2>[A-Z]+)/(?P<tile3>[A-Z]+)/(?P<year>[0-9]+)/(?P<month>[0-9]+)/"
    + r"(?P<day>[0-9]+)/(?P<num>[0-9]+)/auxiliary/(?P<file>AUX_.+)$"
)
# S2 L2A QI Masks
S2_TILE_QI_MSK_REGEX = re.compile(
    r"^tiles/(?P<tile1>[0-9]+)/(?P<tile2>[A-Z]+)/(?P<tile3>[A-Z]+)/(?P<year>[0-9]+)/(?P<month>[0-9]+)/"
    + r"(?P<day>[0-9]+)/(?P<num>[0-9]+)/qi/(?P<file_base>.+)_(?P<file_suffix>[0-9]+m\.jp2)$"
)
# S2 L2A QI PVI
S2_TILE_QI_PVI_REGEX = re.compile(
    r"^tiles/(?P<tile1>[0-9]+)/(?P<tile2>[A-Z]+)/(?P<tile3>[A-Z]+)/(?P<year>[0-9]+)/(?P<month>[0-9]+)/"
    + r"(?P<day>[0-9]+)/(?P<num>[0-9]+)/qi/L2A_PVI\.jp2$"
)
# S2 Tile files ---------------------------------------------------------------
S2_TILE_IMG_REGEX = re.compile(
    r"^tiles/(?P<tile1>[0-9]+)/(?P<tile2>[A-Z]+)/(?P<tile3>[A-Z]+)/(?P<year>[0-9]+)/(?P<month>[0-9]+)/"
    + r"(?P<day>[0-9]+)/(?P<num>[0-9]+)/(?P<file>[A-Z0-9_]+\.jp2)$"
)
S2_TILE_PREVIEW_DIR_REGEX = re.compile(
    r"^tiles/(?P<tile1>[0-9]+)/(?P<tile2>[A-Z]+)/(?P<tile3>[A-Z]+)/(?P<year>[0-9]+)/(?P<month>[0-9]+)/"
    + r"(?P<day>[0-9]+)/(?P<num>[0-9]+)/preview/(?P<file>.+)$"
)
S2_TILE_AUX_DIR_REGEX = re.compile(
    r"^tiles/(?P<tile1>[0-9]+)/(?P<tile2>[A-Z]+)/(?P<tile3>[A-Z]+)/(?P<year>[0-9]+)/(?P<month>[0-9]+)/"
    + r"(?P<day>[0-9]+)/(?P<num>[0-9]+)/auxiliary/(?P<file>.+)$"
)
S2_TILE_QI_DIR_REGEX = re.compile(
    r"^tiles/(?P<tile1>[0-9]+)/(?P<tile2>[A-Z]+)/(?P<tile3>[A-Z]+)/(?P<year>[0-9]+)/(?P<month>[0-9]+)/"
    + r"(?P<day>[0-9]+)/(?P<num>[0-9]+)/qi/(?P<file>.+)$"
)
S2_TILE_THUMBNAIL_REGEX = re.compile(
    r"^tiles/(?P<tile1>[0-9]+)/(?P<tile2>[A-Z]+)/(?P<tile3>[A-Z]+)/(?P<year>[0-9]+)/(?P<month>[0-9]+)/"
    + r"(?P<day>[0-9]+)/(?P<num>[0-9]+)/(?P<file>preview\.\w+)$"
)
S2_TILE_MTD_REGEX = re.compile(
    r"^tiles/(?P<tile1>[0-9]+)/(?P<tile2>[A-Z]+)/(?P<tile3>[A-Z]+)/(?P<year>[0-9]+)/(?P<month>[0-9]+)/"
    + r"(?P<day>[0-9]+)/(?P<num>[0-9]+)/(?P<file>metadata\.xml)$"
)
# S2 Tile generic
S2_TILE_REGEX = re.compile(
    r"^tiles/(?P<tile1>[0-9]+)/(?P<tile2>[A-Z]+)/(?P<tile3>[A-Z]+)/(?P<year>[0-9]+)/(?P<month>[0-9]+)/"
    + r"(?P<day>[0-9]+)/(?P<num>[0-9]+)/(?P<file>.+)$"
)
# S2 Product files
S2_PROD_REGEX = re.compile(
    r"^products/(?P<year>[0-9]+)/(?P<month>[0-9]+)/(?P<day>[0-9]+)/(?P<title>[A-Z0-9_]+)/(?P<file>.+)$"
)
S2_PROD_DS_MTD_REGEX = re.compile(
    r"^products/(?P<year>[0-9]+)/(?P<month>[0-9]+)/(?P<day>[0-9]+)/(?P<title>[A-Z0-9_]+)/datastrip/"
    + r"(?P<num>.+)/(?P<file>metadata\.xml)$"
)
S2_PROD_DS_QI_REGEX = re.compile(
    r"^products/(?P<year>[0-9]+)/(?P<month>[0-9]+)/(?P<day>[0-9]+)/(?P<title>[A-Z0-9_]+)/datastrip/"
    + r"(?P<num>.+)/qi/(?P<file>.+)$"
)
S2_PROD_DS_QI_REPORT_REGEX = re.compile(
    r"^products/(?P<year>[0-9]+)/(?P<month>[0-9]+)/(?P<day>[0-9]+)/(?P<title>[A-Z0-9_]+)/datastrip/"
    + r"(?P<num>.+)/qi/(?P<filename>.+)_report\.xml$"
)
S2_PROD_INSPIRE_REGEX = re.compile(
    r"^products/(?P<year>[0-9]+)/(?P<month>[0-9]+)/(?P<day>[0-9]+)/(?P<title>[A-Z0-9_]+)/"
    + r"(?P<file>inspire\.xml)$"
)
# S2 Product generic
S2_PROD_MTD_REGEX = re.compile(
    r"^products/(?P<year>[0-9]+)/(?P<month>[0-9]+)/(?P<day>[0-9]+)/(?P<title>[A-Z0-9_]+)/"
    + r"(?P<file>metadata\.xml)$"
)
# S1 files --------------------------------------------------------------------
S1_CALIB_REGEX = re.compile(
    r"^GRD/(?P<year>[0-9]+)/(?P<month>[0-9]+)/(?P<day>[0-9]+)/(?P<beam>[A-Z0-9]+)/(?P<prod_pol>[A-Z0-9]+)/"
    + r"(?P<title>[A-Z0-9_]+)/annotation/calibration/"
    + r"(?P<file_prefix>[a-z]+)-(?P<file_beam>[a-z]+)-(?P<file_pol>.+)\.xml$"
)
S1_ANNOT_REGEX = re.compile(
    r"^GRD/(?P<year>[0-9]+)/(?P<month>[0-9]+)/(?P<day>[0-9]+)/(?P<beam>[A-Z0-9]+)/(?P<prod_pol>[A-Z0-9]+)/"
    + r"(?P<title>[A-Z0-9_]+)/annotation/"
    + r"(?P<file_beam>[a-z]+)-(?P<file_pol>.+)\.xml$"
)
S1_MEAS_REGEX = re.compile(
    r"^GRD/(?P<year>[0-9]+)/(?P<month>[0-9]+)/(?P<day>[0-9]+)/(?P<beam>[A-Z0-9]+)/(?P<prod_pol>[A-Z0-9]+)/"
    + r"(?P<title>[A-Z0-9_]+)/measurement/"
    + r"(?P<file_beam>[a-z]+)-(?P<file_pol>.+)\.(?P<file_ext>[a-z0-9]+)$"
)
S1_REPORT_REGEX = re.compile(
    r"^GRD/(?P<year>[0-9]+)/(?P<month>[0-9]+)/(?P<day>[0-9]+)/(?P<beam>[A-Z0-9]+)/(?P<prod_pol>[A-Z0-9]+)/"
    + r"(?P<title>[A-Z0-9_]+)/(?P<file>report-\w+\.pdf)$"
)
# S1 generic
S1_REGEX = re.compile(
    r"^GRD/[0-9]{4}/[0-9]+/[0-9]+/[A-Z0-9]+/[A-Z0-9]+/(?P<title>S1[A-Z0-9_]+)/(?P<file>.+)$"
)
# CBERS4 generic
CBERS4_REGEX = re.compile(
    r"^GRD/[0-9]{4}/[0-9]+/[0-9]+/[A-Z0-9]+/[A-Z0-9]+/(?P<title>S1[A-Z0-9_]+)/(?P<file>.+)$"
)

# S1 image number conf per polarization ---------------------------------------
S1_IMG_NB_PER_POLAR = {
    "SH": {"HH": 1},
    "SV": {"VV": 1},
    "DH": {"HH": 1, "HV": 2},
    "DV": {"VV": 1, "VH": 2},
    "HH": {"HH": 1},
    "HV": {"HV": 1},
    "VV": {"VV": 1},
    "VH": {"VH": 1},
}


class AwsDownload(Download):
    """Download on AWS using S3 protocol.

    :param provider: provider name
    :type provider: str
    :param config: Download plugin configuration
                   config.requester_pays:

    :param config: Download plugin configuration:

                     * ``config.base_uri`` (str) - s3 endpoint url
                     * ``config.requester_pays`` (bool) - whether download is done from
                       a requester-pays bucket or not
                     * ``config.flatten_top_dirs`` (bool) - flatten directory structure
                     * ``config.products`` (dict) - product_type specific configuration
    :type config: :class:`~eodag.config.PluginConfig`
    """

    def __init__(self, provider, config):
        super(AwsDownload, self).__init__(provider, config)
        self.requester_pays = getattr(self.config, "requester_pays", False)
        self.s3_session = None

    def download(self, product, auth=None, progress_callback=None, **kwargs):
        """Download method for AWS S3 API.

        The product can be downloaded as it is, or as SAFE-formatted product.
        SAFE-build is configured for a given provider and product type.
        If the product title is configured to be updated during download and
        SAFE-formatted, its destination path will be:
        `{outputs_prefix}/{title}/{updated_title}.SAFE`

        :param product: The EO product to download
        :type product: :class:`~eodag.api.product._product.EOProduct`
        :param auth: (optional) The configuration of a plugin of type Authentication
        :type auth: :class:`~eodag.config.PluginConfig`
        :param progress_callback: (optional) A method or a callable object
                                  which takes a current size and a maximum
                                  size as inputs and handle progress bar
                                  creation and update to give the user a
                                  feedback on the download progress
        :type progress_callback: :class:`~eodag.utils.ProgressCallback` or None
        :param kwargs: `outputs_prefix` (str), `extract` (bool), `delete_archive` (bool)
                        and `dl_url_params` (dict) can be provided as additional kwargs
                        and will override any other values defined in a configuration
                        file or with environment variables.
        :type kwargs: Union[str, bool, dict]
        :returns: The absolute path to the downloaded product in the local filesystem
        :rtype: str
        """
        if progress_callback is None:
            logger.info(
                "Progress bar unavailable, please call product.download() instead of plugin.download()"
            )
            progress_callback = ProgressCallback(disable=True)

        # prepare download & create dirs (before updating metadata)
        product_local_path, record_filename = self._prepare_download(
            product, progress_callback=progress_callback, **kwargs
        )
        if not product_local_path or not record_filename:
            if product_local_path:
                product.location = path_to_uri(product_local_path)
            return product_local_path
        product_local_path = product_local_path.replace(".zip", "")
        # remove existing incomplete file
        if os.path.isfile(product_local_path):
            os.remove(product_local_path)
        # create product dest dir
        if not os.path.isdir(product_local_path):
            os.makedirs(product_local_path)

        product_conf = getattr(self.config, "products", {}).get(
            product.product_type, {}
        )

        build_safe = product_conf.get("build_safe", False)

        # product conf overrides provider conf for "flatten_top_dirs"
        flatten_top_dirs = product_conf.get(
            "flatten_top_dirs", getattr(self.config, "flatten_top_dirs", False)
        )

        # xtra metadata needed for SAFE product
        if build_safe and "fetch_metadata" in product_conf.keys():
            fetch_format = product_conf["fetch_metadata"]["fetch_format"]
            update_metadata = product_conf["fetch_metadata"]["update_metadata"]
            fetch_url = product_conf["fetch_metadata"]["fetch_url"].format(
                **product.properties
            )
            logger.info("Fetching extra metadata from %s" % fetch_url)
            resp = requests.get(fetch_url, timeout=HTTP_REQ_TIMEOUT)
            update_metadata = mtd_cfg_as_jsonpath(update_metadata)
            if fetch_format == "json":
                json_resp = resp.json()
                update_metadata = properties_from_json(json_resp, update_metadata)
                product.properties.update(update_metadata)
            elif fetch_format == "xml":
                update_metadata = properties_from_xml(resp.content, update_metadata)
                product.properties.update(update_metadata)
            else:
                logger.warning(
                    "SAFE metadata fetch format %s not implemented" % fetch_format
                )
        # if assets are defined, use them instead of scanning product.location
        if hasattr(product, "assets") and not getattr(
            self.config, "ignore_assets", False
        ):
            bucket_names_and_prefixes = []
            for complementary_url in getattr(product, "assets", {}).values():
                bucket_names_and_prefixes.append(
                    self.get_bucket_name_and_prefix(
                        product, complementary_url.get("href", "")
                    )
                )
        else:
            bucket_names_and_prefixes = [self.get_bucket_name_and_prefix(product)]

        # add complementary urls
        try:
            for complementary_url_key in product_conf.get("complementary_url_key", []):
                bucket_names_and_prefixes.append(
                    self.get_bucket_name_and_prefix(
                        product, product.properties[complementary_url_key]
                    )
                )
        except KeyError:
            logger.error(
                "complementary_url_key %s is missing in %s properties"
                % (complementary_url_key, product.properties["id"])
            )

        # authenticate
        authenticated_objects = {}
        auth_error_messages = set()
        for idx, pack in enumerate(bucket_names_and_prefixes):
            try:
                bucket_name, prefix = pack
                if bucket_name not in authenticated_objects:
                    # get Prefixes longest common base path
                    common_prefix = ""
                    prefix_split = prefix.split("/")
                    prefixes_in_bucket = len(
                        [p for b, p in bucket_names_and_prefixes if b == bucket_name]
                    )
                    for i in range(1, len(prefix_split)):
                        common_prefix = "/".join(prefix_split[0:i])
                        if (
                            len(
                                [
                                    p
                                    for b, p in bucket_names_and_prefixes
                                    if b == bucket_name and common_prefix in p
                                ]
                            )
                            < prefixes_in_bucket
                        ):
                            common_prefix = "/".join(prefix_split[0 : i - 1])
                            break
                    # connect to aws s3 and get bucket auhenticated objects
                    s3_objects = self.get_authenticated_objects(
                        bucket_name, common_prefix, auth
                    )
                    authenticated_objects[bucket_name] = s3_objects
                else:
                    s3_objects = authenticated_objects[bucket_name]

            except AuthenticationError as e:
                logger.warning("Unexpected error: %s" % e)
                logger.warning("Skipping %s/%s" % (bucket_name, prefix))
                auth_error_messages.add(str(e))
            except ClientError as e:
                err = e.response["Error"]
                auth_messages = [
                    "AccessDenied",
                    "InvalidAccessKeyId",
                    "SignatureDoesNotMatch",
                ]
                if err["Code"] in auth_messages and "key" in err["Message"].lower():
                    raise AuthenticationError(
                        "HTTP error {} returned\n{}: {}\nPlease check your credentials for {}".format(
                            e.response["ResponseMetadata"]["HTTPStatusCode"],
                            err["Code"],
                            err["Message"],
                            self.provider,
                        )
                    )
                logger.warning("Unexpected error: %s" % e)
                logger.warning("Skipping %s/%s" % (bucket_name, prefix))
                auth_error_messages.add(str(e))

        # could not auth on any bucket
        if not authenticated_objects:
            raise AuthenticationError(", ".join(auth_error_messages))

        # downloadable files
        product_chunks = []
        for bucket_name, prefix in bucket_names_and_prefixes:
            # unauthenticated items filtered out
            if bucket_name in authenticated_objects.keys():
                product_chunks.extend(
                    authenticated_objects[bucket_name].filter(Prefix=prefix)
                )

        unique_product_chunks = set(product_chunks)

        total_size = sum([p.size for p in unique_product_chunks])

        # download
        progress_callback.reset(total=total_size)
        try:
            for product_chunk in unique_product_chunks:
                chunk_rel_path = self.get_chunk_dest_path(
                    product,
                    product_chunk,
                    build_safe=build_safe,
                    dir_prefix=prefix,
                )
                chunk_abs_path = os.path.join(product_local_path, chunk_rel_path)
                chunk_abs_path_dir = os.path.dirname(chunk_abs_path)
                if not os.path.isdir(chunk_abs_path_dir):
                    os.makedirs(chunk_abs_path_dir)

                if not os.path.isfile(chunk_abs_path):
                    product_chunk.Bucket().download_file(
                        product_chunk.key,
                        chunk_abs_path,
                        ExtraArgs=getattr(s3_objects, "_params", {}),
                        Callback=progress_callback,
                    )

        except AuthenticationError as e:
            logger.warning("Unexpected error: %s" % e)
            logger.warning("Skipping %s/%s" % (bucket_name, prefix))
        except ClientError as e:
            err = e.response["Error"]
            auth_messages = [
                "AccessDenied",
                "InvalidAccessKeyId",
                "SignatureDoesNotMatch",
            ]
            if err["Code"] in auth_messages and "key" in err["Message"].lower():
                raise AuthenticationError(
                    "HTTP error {} returned\n{}: {}\nPlease check your credentials for {}".format(
                        e.response["ResponseMetadata"]["HTTPStatusCode"],
                        err["Code"],
                        err["Message"],
                        self.provider,
                    )
                )
            logger.warning("Unexpected error: %s" % e)
            logger.warning("Skipping %s/%s" % (bucket_name, prefix))

        # finalize safe product
        if build_safe and "S2_MSI" in product.product_type:
            self.finalize_s2_safe_product(product_local_path)
        # flatten directory structure
        elif flatten_top_dirs:
            tmp_product_local_path = "%s-tmp" % product_local_path
            for d, dirs, files in os.walk(product_local_path):
                if len(files) != 0:
                    shutil.copytree(d, tmp_product_local_path)
                    shutil.rmtree(product_local_path)
                    os.rename(tmp_product_local_path, product_local_path)
                    break

        if build_safe:
            self.check_manifest_file_list(product_local_path)

        # save hash/record file
        with open(record_filename, "w") as fh:
            fh.write(product.remote_location)
        logger.debug("Download recorded in %s", record_filename)

        product.location = path_to_uri(product_local_path)
        return product_local_path

    def get_rio_env(self, bucket_name, prefix, auth_dict):
        """Get rasterio environment variables needed for data access authentication.

        :param bucket_name: Bucket containg objects
        :type bucket_name: str
        :param prefix: Prefix used to try auth
        :type prefix: str
        :param auth_dict: Dictionnary containing authentication keys
        :type auth_dict: dict
        :returns: The rasterio environement variables
        :rtype: dict
        """
        if self.s3_session is not None:
            if self.requester_pays:
                return {"session": self.s3_session, "requester_pays": True}
            else:
                return {"session": self.s3_session}

        _ = self.get_authenticated_objects(bucket_name, prefix, auth_dict)
        if self.s3_session is not None:
            if self.requester_pays:
                return {"session": self.s3_session, "requester_pays": True}
            else:
                return {"session": self.s3_session}
        else:
            return {"aws_unsigned": True}

    def get_authenticated_objects(self, bucket_name, prefix, auth_dict):
        """Get boto3 authenticated objects for the given bucket using
        the most adapted auth strategy.
        Also expose ``s3_session`` as class variable if available.

        :param bucket_name: Bucket containg objects
        :type bucket_name: str
        :param prefix: Prefix used to filter objects on auth try
                       (not used to filter returned objects)
        :type prefix: str
        :param auth_dict: Dictionnary containing authentication keys
        :type auth_dict: dict
        :returns: The boto3 authenticated objects
        :rtype: :class:`~boto3.resources.collection.s3.Bucket.objectsCollection`
        """
        auth_methods = [
            self._get_authenticated_objects_unsigned,
            self._get_authenticated_objects_from_auth_profile,
            self._get_authenticated_objects_from_auth_keys,
            self._get_authenticated_objects_from_env,
        ]
        # skip _get_authenticated_objects_from_env if credentials were filled in eodag conf
        if auth_dict:
            del auth_methods[-1]

        for try_auth_method in auth_methods:
            try:
                s3_objects = try_auth_method(bucket_name, prefix, auth_dict)
                if s3_objects:
                    logger.debug("Auth using %s succeeded", try_auth_method.__name__)
                    return s3_objects
            except ClientError as e:
                if e.response.get("Error", {}).get("Code", {}) in [
                    "AccessDenied",
                    "InvalidAccessKeyId",
                    "SignatureDoesNotMatch",
                ]:
                    pass
                else:
                    raise e
            except ProfileNotFound:
                pass
            logger.debug("Auth using %s failed", try_auth_method.__name__)

        raise AuthenticationError(
            "Unable do authenticate on s3://%s using any available credendials configuration"
            % bucket_name
        )

    def _get_authenticated_objects_unsigned(self, bucket_name, prefix, auth_dict):
        """Auth strategy using no-sign-request"""

        s3_resource = boto3.resource(
            service_name="s3", endpoint_url=getattr(self.config, "base_uri", None)
        )
        s3_resource.meta.client.meta.events.register(
            "choose-signer.s3.*", disable_signing
        )
        objects = s3_resource.Bucket(bucket_name).objects
        list(objects.filter(Prefix=prefix).limit(1))
        return objects

    def _get_authenticated_objects_from_auth_profile(
        self, bucket_name, prefix, auth_dict
    ):
        """Auth strategy using RequestPayer=requester and ``aws_profile`` from provided credentials"""

        if "profile_name" in auth_dict.keys():
            s3_session = boto3.session.Session(profile_name=auth_dict["profile_name"])
            s3_resource = s3_session.resource(
                service_name="s3",
                endpoint_url=getattr(self.config, "base_uri", None),
            )
            if self.requester_pays:
                objects = s3_resource.Bucket(bucket_name).objects.filter(
                    RequestPayer="requester"
                )
            else:
                objects = s3_resource.Bucket(bucket_name).objects
            list(objects.filter(Prefix=prefix).limit(1))
            self.s3_session = s3_session
            return objects
        else:
            return None

    def _get_authenticated_objects_from_auth_keys(self, bucket_name, prefix, auth_dict):
        """Auth strategy using RequestPayer=requester and ``aws_access_key_id``/``aws_secret_access_key``
        from provided credentials"""

        if all(k in auth_dict for k in ("aws_access_key_id", "aws_secret_access_key")):
            s3_session = boto3.session.Session(
                aws_access_key_id=auth_dict["aws_access_key_id"],
                aws_secret_access_key=auth_dict["aws_secret_access_key"],
            )
            s3_resource = s3_session.resource(
                service_name="s3",
                endpoint_url=getattr(self.config, "base_uri", None),
            )
            if self.requester_pays:
                objects = s3_resource.Bucket(bucket_name).objects.filter(
                    RequestPayer="requester"
                )
            else:
                objects = s3_resource.Bucket(bucket_name).objects
            list(objects.filter(Prefix=prefix).limit(1))
            self.s3_session = s3_session
            return objects
        else:
            return None

    def _get_authenticated_objects_from_env(self, bucket_name, prefix, auth_dict):
        """Auth strategy using RequestPayer=requester and current environment"""

        s3_session = boto3.session.Session()
        s3_resource = s3_session.resource(
            service_name="s3", endpoint_url=getattr(self.config, "base_uri", None)
        )
        if self.requester_pays:
            objects = s3_resource.Bucket(bucket_name).objects.filter(
                RequestPayer="requester"
            )
        else:
            objects = s3_resource.Bucket(bucket_name).objects
        list(objects.filter(Prefix=prefix).limit(1))
        self.s3_session = s3_session
        return objects

    def get_bucket_name_and_prefix(self, product, url=None):
        """Extract bucket name and prefix from product URL

        :param product: The EO product to download
        :type product: :class:`~eodag.api.product._product.EOProduct`
        :param url: (optional) URL to use as product.location
        :type url: str
        :returns: bucket_name and prefix as str
        :rtype: tuple
        """
        if not url:
            url = product.location
        bucket, prefix = None, None
        # Assume the bucket is encoded into the product location as a URL or given as the 'default_bucket' config
        # param
        scheme, netloc, path, params, query, fragment = urlparse(url)
        if not scheme or scheme == "s3":
            bucket = (
                netloc
                if netloc
                else getattr(self.config, "products", {})
                .get(product.product_type, {})
                .get("default_bucket", "")
            )
            prefix = path.strip("/")
        elif scheme in ("http", "https", "ftp"):
            parts = path.split("/")
            bucket, prefix = parts[1], "/{}".format("/".join(parts[2:]))
        return bucket, prefix

    def check_manifest_file_list(self, product_path):
        """Checks if products listed in manifest.safe exist"""
        manifest_path_list = [
            os.path.join(d, x)
            for d, _, f in os.walk(product_path)
            for x in f
            if x == "manifest.safe"
        ]
        if len(manifest_path_list) == 0:
            raise FileNotFoundError(
                f"No manifest.safe could be found in {product_path}"
            )
        else:
            safe_path = os.path.dirname(manifest_path_list[0])

        root = etree.parse(os.path.join(safe_path, "manifest.safe")).getroot()
        for safe_file in root.xpath("//fileLocation"):
            safe_file_path = os.path.join(safe_path, safe_file.get("href"))
            if not os.path.isfile(safe_file_path) and "HTML" in safe_file.get("href"):
                # add empty files for missing HTML/*
                Path(safe_file_path).touch()
            elif not os.path.isfile(safe_file_path):
                logger.warning("SAFE build: %s is missing" % safe_file.get("href"))

    def finalize_s2_safe_product(self, product_path):
        """Add missing dirs to downloaded product"""
        try:
            logger.debug("Finalize SAFE product")
            manifest_path_list = [
                os.path.join(d, x)
                for d, _, f in os.walk(product_path)
                for x in f
                if x == "manifest.safe"
            ]
            if len(manifest_path_list) == 0:
                raise FileNotFoundError(
                    f"No manifest.safe could be found in {product_path}"
                )
            else:
                safe_path = os.path.dirname(manifest_path_list[0])

            # create empty missing dirs
            auxdata_path = os.path.join(safe_path, "AUX_DATA")
            if not os.path.isdir(auxdata_path):
                os.makedirs(auxdata_path)
            html_path = os.path.join(safe_path, "HTML")
            if not os.path.isdir(html_path):
                os.makedirs(html_path)
            repinfo_path = os.path.join(safe_path, "rep_info")
            if not os.path.isdir(repinfo_path):
                os.makedirs(repinfo_path)

            # granule tile dirname
            root = etree.parse(os.path.join(safe_path, "manifest.safe")).getroot()
            tile_id = os.path.basename(
                os.path.dirname(
                    root.xpath("//fileLocation[contains(@href,'MTD_TL.xml')]")[0].get(
                        "href"
                    )
                )
            )
            granule_folder = os.path.join(safe_path, "GRANULE")
            rename_subfolder(granule_folder, tile_id)

            # datastrip scene dirname
            scene_id = os.path.basename(
                os.path.dirname(
                    root.xpath("//fileLocation[contains(@href,'MTD_DS.xml')]")[0].get(
                        "href"
                    )
                )
            )
            datastrip_folder = os.path.join(safe_path, "DATASTRIP")
            rename_subfolder(datastrip_folder, scene_id)
        except Exception as e:
            logger.exception("Could not finalize SAFE product from downloaded data")
            raise DownloadError(e)

    def get_chunk_dest_path(self, product, chunk, dir_prefix, build_safe=False):
        """Get chunk destination path"""
        if build_safe:
            # S2 common
            if "S2_MSI" in product.product_type:
                title_search = re.search(
                    r"^\w+_\w+_(\w+)_(\w+)_(\w+)_(\w+)_(\w+)$",
                    product.properties["title"],
                )
                title_date1 = title_search.group(1) if title_search else None
                title_part3 = title_search.group(4) if title_search else None
                ds_dir_search = re.search(
                    r"^.+_(DS_\w+_+\w+_\w+)_\w+.\w+$",
                    product.properties.get("originalSceneID", ""),
                )
                ds_dir = ds_dir_search.group(1) if ds_dir_search else 0
                s2_processing_level = product.product_type.split("_")[-1]
            # S1 common
            elif product.product_type == "S1_SAR_GRD":
                s1_title_suffix_search = re.search(
                    r"^.+_([A-Z0-9_]+_[A-Z0-9_]+_[A-Z0-9_]+_[A-Z0-9_]+)_\w+$",
                    product.properties["title"],
                )
                s1_title_suffix = (
                    s1_title_suffix_search.group(1).lower().replace("_", "-")
                    if s1_title_suffix_search
                    else None
                )

            # S2 L2A Tile files -----------------------------------------------
            if S2L2A_TILE_IMG_REGEX.match(chunk.key):
                found_dict = S2L2A_TILE_IMG_REGEX.match(chunk.key).groupdict()
                product_path = (
                    "%s.SAFE/GRANULE/%s/IMG_DATA/R%s/T%s%s%s_%s_%s_%s.jp2"
                    % (
                        product.properties["title"],
                        found_dict["num"],
                        found_dict["res"],
                        found_dict["tile1"],
                        found_dict["tile2"],
                        found_dict["tile3"],
                        title_date1,
                        found_dict["file"],
                        found_dict["res"],
                    )
                )
            elif S2L2A_TILE_AUX_DIR_REGEX.match(chunk.key):
                found_dict = S2L2A_TILE_AUX_DIR_REGEX.match(chunk.key).groupdict()
                product_path = "%s.SAFE/GRANULE/%s/AUX_DATA/%s" % (
                    product.properties["title"],
                    found_dict["num"],
                    found_dict["file"],
                )
            # S2 L2A QI Masks
            elif S2_TILE_QI_MSK_REGEX.match(chunk.key):
                found_dict = S2_TILE_QI_MSK_REGEX.match(chunk.key).groupdict()
                product_path = "%s.SAFE/GRANULE/%s/QI_DATA/MSK_%sPRB_%s" % (
                    product.properties["title"],
                    found_dict["num"],
                    found_dict["file_base"],
                    found_dict["file_suffix"],
                )
            # S2 L2A QI PVI
            elif S2_TILE_QI_PVI_REGEX.match(chunk.key):
                found_dict = S2_TILE_QI_PVI_REGEX.match(chunk.key).groupdict()
                product_path = "%s.SAFE/GRANULE/%s/QI_DATA/%s_%s_PVI.jp2" % (
                    product.properties["title"],
                    found_dict["num"],
                    title_part3,
                    title_date1,
                )
            # S2 Tile files ---------------------------------------------------
            elif S2_TILE_PREVIEW_DIR_REGEX.match(chunk.key):
                found_dict = S2_TILE_PREVIEW_DIR_REGEX.match(chunk.key).groupdict()
                product_path = "%s.SAFE/GRANULE/%s/preview/%s" % (
                    product.properties["title"],
                    found_dict["num"],
                    found_dict["file"],
                )
            elif S2_TILE_IMG_REGEX.match(chunk.key):
                found_dict = S2_TILE_IMG_REGEX.match(chunk.key).groupdict()
                product_path = "%s.SAFE/GRANULE/%s/IMG_DATA/T%s%s%s_%s_%s" % (
                    product.properties["title"],
                    found_dict["num"],
                    found_dict["tile1"],
                    found_dict["tile2"],
                    found_dict["tile3"],
                    title_date1,
                    found_dict["file"],
                )
            elif S2_TILE_THUMBNAIL_REGEX.match(chunk.key):
                found_dict = S2_TILE_THUMBNAIL_REGEX.match(chunk.key).groupdict()
                product_path = "%s.SAFE/GRANULE/%s/%s" % (
                    product.properties["title"],
                    found_dict["num"],
                    found_dict["file"],
                )
            elif S2_TILE_MTD_REGEX.match(chunk.key):
                found_dict = S2_TILE_MTD_REGEX.match(chunk.key).groupdict()
                product_path = "%s.SAFE/GRANULE/%s/MTD_TL.xml" % (
                    product.properties["title"],
                    found_dict["num"],
                )
            elif S2_TILE_AUX_DIR_REGEX.match(chunk.key):
                found_dict = S2_TILE_AUX_DIR_REGEX.match(chunk.key).groupdict()
                product_path = "%s.SAFE/GRANULE/%s/AUX_DATA/AUX_%s" % (
                    product.properties["title"],
                    found_dict["num"],
                    found_dict["file"],
                )
            elif S2_TILE_QI_DIR_REGEX.match(chunk.key):
                found_dict = S2_TILE_QI_DIR_REGEX.match(chunk.key).groupdict()
                product_path = "%s.SAFE/GRANULE/%s/QI_DATA/%s" % (
                    product.properties["title"],
                    found_dict["num"],
                    found_dict["file"],
                )
            # S2 Tiles generic
            elif S2_TILE_REGEX.match(chunk.key):
                found_dict = S2_TILE_REGEX.match(chunk.key).groupdict()
                product_path = "%s.SAFE/GRANULE/%s/%s" % (
                    product.properties["title"],
                    found_dict["num"],
                    found_dict["file"],
                )
            # S2 Product files
            elif S2_PROD_DS_MTD_REGEX.match(chunk.key):
                found_dict = S2_PROD_DS_MTD_REGEX.match(chunk.key).groupdict()
                product_path = "%s.SAFE/DATASTRIP/%s/MTD_DS.xml" % (
                    product.properties["title"],
                    ds_dir,
                )
            elif S2_PROD_DS_QI_REPORT_REGEX.match(chunk.key):
                found_dict = S2_PROD_DS_QI_REPORT_REGEX.match(chunk.key).groupdict()
                product_path = "%s.SAFE/DATASTRIP/%s/QI_DATA/%s.xml" % (
                    product.properties["title"],
                    ds_dir,
                    found_dict["filename"],
                )
            elif S2_PROD_DS_QI_REGEX.match(chunk.key):
                found_dict = S2_PROD_DS_QI_REGEX.match(chunk.key).groupdict()
                product_path = "%s.SAFE/DATASTRIP/%s/QI_DATA/%s" % (
                    product.properties["title"],
                    ds_dir,
                    found_dict["file"],
                )
            elif S2_PROD_INSPIRE_REGEX.match(chunk.key):
                found_dict = S2_PROD_INSPIRE_REGEX.match(chunk.key).groupdict()
                product_path = "%s.SAFE/INSPIRE.xml" % (product.properties["title"],)
            elif S2_PROD_MTD_REGEX.match(chunk.key):
                found_dict = S2_PROD_MTD_REGEX.match(chunk.key).groupdict()
                product_path = "%s.SAFE/MTD_MSI%s.xml" % (
                    product.properties["title"],
                    s2_processing_level,
                )
            # S2 Product generic
            elif S2_PROD_REGEX.match(chunk.key):
                found_dict = S2_PROD_REGEX.match(chunk.key).groupdict()
                product_path = "%s.SAFE/%s" % (
                    product.properties["title"],
                    found_dict["file"],
                )
            # S1 --------------------------------------------------------------
            elif S1_CALIB_REGEX.match(chunk.key):
                found_dict = S1_CALIB_REGEX.match(chunk.key).groupdict()
                product_path = (
                    "%s.SAFE/annotation/calibration/%s-%s-%s-grd-%s-%s-%03d.xml"
                    % (
                        product.properties["title"],
                        found_dict["file_prefix"],
                        product.properties["platformSerialIdentifier"].lower(),
                        found_dict["file_beam"],
                        found_dict["file_pol"],
                        s1_title_suffix,
                        S1_IMG_NB_PER_POLAR.get(
                            product.properties["polarizationMode"], {}
                        ).get(found_dict["file_pol"].upper(), 1),
                    )
                )
            elif S1_ANNOT_REGEX.match(chunk.key):
                found_dict = S1_ANNOT_REGEX.match(chunk.key).groupdict()
                product_path = "%s.SAFE/annotation/%s-%s-grd-%s-%s-%03d.xml" % (
                    product.properties["title"],
                    product.properties["platformSerialIdentifier"].lower(),
                    found_dict["file_beam"],
                    found_dict["file_pol"],
                    s1_title_suffix,
                    S1_IMG_NB_PER_POLAR.get(
                        product.properties["polarizationMode"], {}
                    ).get(found_dict["file_pol"].upper(), 1),
                )
            elif S1_MEAS_REGEX.match(chunk.key):
                found_dict = S1_MEAS_REGEX.match(chunk.key).groupdict()
                product_path = "%s.SAFE/measurement/%s-%s-grd-%s-%s-%03d.%s" % (
                    product.properties["title"],
                    product.properties["platformSerialIdentifier"].lower(),
                    found_dict["file_beam"],
                    found_dict["file_pol"],
                    s1_title_suffix,
                    S1_IMG_NB_PER_POLAR.get(
                        product.properties["polarizationMode"], {}
                    ).get(found_dict["file_pol"].upper(), 1),
                    found_dict["file_ext"],
                )
            elif S1_REPORT_REGEX.match(chunk.key):
                found_dict = S1_REPORT_REGEX.match(chunk.key).groupdict()
                product_path = "%s.SAFE/%s.SAFE-%s" % (
                    product.properties["title"],
                    product.properties["title"],
                    found_dict["file"],
                )
            # S1 generic
            elif S1_REGEX.match(chunk.key):
                found_dict = S1_REGEX.match(chunk.key).groupdict()
                product_path = "%s.SAFE/%s" % (
                    product.properties["title"],
                    found_dict["file"],
                )
        # no SAFE format
        else:
            product_path = chunk.key.split(dir_prefix.strip("/") + "/")[-1]
        logger.debug("Downloading %s to %s" % (chunk.key, product_path))
        return product_path

    def download_all(
        self,
        products,
        auth=None,
        downloaded_callback=None,
        progress_callback=None,
        **kwargs,
    ):
        """
        download_all using parent (base plugin) method
        """
        return super(AwsDownload, self).download_all(
            products,
            auth=auth,
            downloaded_callback=downloaded_callback,
            progress_callback=progress_callback,
            **kwargs,
        )



# -*- coding: utf-8 -*-
# Copyright 2018, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from eodag.plugins.base import PluginTopic
from eodag.utils.exceptions import MisconfiguredError


class Authentication(PluginTopic):
    """Plugins authentication Base plugin"""

    def authenticate(self):
        """Authenticate"""
        raise NotImplementedError

    def validate_config_credentials(self):
        """Validate configured credentials"""
        # No credentials dict in the config
        try:
            credentials = self.config.credentials
        except AttributeError:
            raise MisconfiguredError(
                f"Missing credentials configuration for provider {self.provider}"
            )
        # Empty credentials dict
        if not credentials:
            raise MisconfiguredError(
                f"Missing credentials for provider {self.provider}"
            )
        # Credentials keys but values are None.
        missing_credentials = [
            cred_name
            for cred_name, cred_value in credentials.items()
            if cred_value is None
        ]
        if missing_credentials:
            raise MisconfiguredError(
                "The following credentials are missing for provider {}: {}".format(
                    self.provider, ", ".join(missing_credentials)
                )
            )
